@@ -83,21 +83,32 @@ gdip_bitmap_init (GpBitmap *bitmap)
 	bitmap->data.Left = 0;
 	bitmap->data.Top = 0;
 	bitmap->cairo_format = CAIRO_FORMAT_ARGB32; 
-		
-	bitmap->hBitmapDC = 0;
-	bitmap->hInitialBitmap = 0;
-	bitmap->hBitmap = 0;
 }
 
 
 
-void
+GpStatus
 gdip_bitmap_clone (GpBitmap *bitmap, GpBitmap **clonedbitmap)
 {
 	GpBitmap *result = (GpBitmap *) GdipAlloc (sizeof (GpBitmap));	
+	if (result == NULL) {
+		return OutOfMemory;
+	}
 	memcpy (result, bitmap, sizeof (GpBitmap));
+
+	result->image.surface = NULL;
+	result->image.palette = NULL;
+	result->image.propItems = NULL;
+	result->image.frameDimensionCount = 0;
+	result->image.frameDimensionList = NULL;
+	result->data.Bytes = NULL;
 	
 	result->data.Scan0 = GdipAlloc (bitmap->data.Stride * bitmap->data.Height);
+	if (result->data.Scan0 == NULL) {
+		GdipFree(result);
+		return OutOfMemory;
+	}
+
 	memcpy (result->data.Scan0, bitmap->data.Scan0, bitmap->data.Stride * bitmap->data.Height);
 	*clonedbitmap = result;
 	result->data.Reserved =  GBD_OWN_SCAN0; /* Also overwrites a possible GDB_LOCKED */
@@ -112,6 +123,8 @@ gdip_bitmap_clone (GpBitmap *bitmap, GpBitmap **clonedbitmap)
 	
 	
 	/*TODO: We should also copy palette info when we support it*/
+
+	return Ok;
 }
 
 
@@ -194,6 +207,9 @@ GdipCreateBitmapFromScan0 (int width, int height, int stride, int format, void *
         bool own_scan0 = FALSE;
 	int cairo_format = 0;
 	
+	if (width <= 0 || height <= 0)
+		return InvalidParameter;
+	
 	switch (format) {
 	case Format24bppRgb:
 		cairo_format = CAIRO_FORMAT_RGB24;  
@@ -368,13 +384,14 @@ GdipCreateBitmapFromGraphics (int width, int height, GpGraphics *graphics, GpBit
 GpStatus
 GdipCreateBitmapFromHBITMAP(void *hbm, void *hpal, GpBitmap** bitmap)
 {
-	return(NotImplemented);
+	return GdipCloneImage((GpImage *)hbm, (GpImage **)bitmap);
 }
 
 GpStatus
 GdipCreateHBITMAPFromBitmap(GpBitmap* bitmap, void **hbmReturn, unsigned long background)
 {
-	return(NotImplemented);
+	*hbmReturn = bitmap;
+	return Ok;
 }
 
 GpStatus
@@ -400,7 +417,6 @@ GdipCloneBitmapAreaI (int x, int y, int width, int height, PixelFormat format,
 					  GpBitmap *original, GpBitmap **bitmap)
 {
 	GpBitmap *result = 0;
-	GdipBitmapData bd;
 	Rect sr = { x, y, width, height };
 	Rect dr = { 0, 0, width, height };
 	GpStatus st;
@@ -411,21 +427,17 @@ GdipCloneBitmapAreaI (int x, int y, int width, int height, PixelFormat format,
 	g_return_val_if_fail (x + width <= original->data.Width, InvalidParameter);
 	g_return_val_if_fail (y + height <= original->data.Height, InvalidParameter);
 
-	bd.Scan0 = NULL;
-	bd.PixelFormat = format;
-	
-	st = gdip_bitmap_clone_data_rect (&original->data, &sr, &bd, &dr);
-	if (st != Ok)
-		return st;
-
 	result = gdip_bitmap_new ();
 	if (result == NULL){
-		GdipFree (bd.Scan0);
 		return OutOfMemory;
 	}
-	
+
+	st = gdip_bitmap_clone_data_rect (&original->data, &sr, &result->data, &dr);
+	if (st != Ok) {
+		return st;
+	}
+
 	result->cairo_format = original->cairo_format;
-	memcpy (&result->data, &bd, sizeof (GdipBitmapData));
 	result->image.pixFormat = format;
 	result->image.format = original->image.format;
 	result->image.height = result->data.Height;
@@ -701,8 +713,12 @@ gdip_init_pixel_stream (StreamingState *state, BitmapData *data, int x, int y, i
 		case Format1bppIndexed: state->one_pixel_mask = 0x01; state->one_pixel_shift = 1; state->pixels_per_byte = 8; break;
 		case Format4bppIndexed: state->one_pixel_mask = 0x0F; state->one_pixel_shift = 4; state->pixels_per_byte = 2; break;
 		case Format8bppIndexed: state->one_pixel_mask = 0xFF; state->one_pixel_shift = 8; state->pixels_per_byte = 1; break;
+		case Format24bppRgb: /* gdip_get_pixel_format_bpp lies because other code depends on Format24bppRgb being CAIRO_FORMAT_ARGB32 internally */
+			state->pixels_per_byte = -3;
+			break;
 		default:
-			state->pixels_per_byte = 0; /* indicate full RGB processing */
+			/* indicate full RGB processing */
+			state->pixels_per_byte = -gdip_get_pixel_format_bpp (data->PixelFormat) / 8; 
 			break;
 	}
 
@@ -811,6 +827,12 @@ gdip_pixel_stream_get_next (StreamingState *state)
 		 * CAIRO_FORMAT_ARGB. This makes this section very easy to implement. If native
 		 * support for 15- and 16-bit pixel formats needs to be added in the future, though,
 		 * then this is where it needs to be done.
+		 *
+		 * In order to simplify advancing the state->scan pointer, the state->pixels_per_byte
+		 * member is set to the number of bytes per pixel, negated. That is, for 24-bit
+		 * formats, it is set to -3, and for 32-bit formats, it is set to -4.
+		 *
+		 * Note that pixel streams do not support 48- and 64-bit data at this time.
 		 */
 		ret = *(unsigned int *)state->scan;
 
@@ -825,7 +847,7 @@ gdip_pixel_stream_get_next (StreamingState *state)
 			ret |= force_alpha;
 		}
 
-		state->scan += 4;
+		state->scan -= state->pixels_per_byte;
 		state->x++;
 
 		if (state->x >= (state->region.X + state->region.Width)) {
@@ -834,7 +856,7 @@ gdip_pixel_stream_get_next (StreamingState *state)
 
 			state->scan = (unsigned char *)(state->data->Scan0)
 			            + state->y * state->data->Stride
-			            + state->x * 4;
+			            + state->x * -state->pixels_per_byte;
 		}
 	}
 
@@ -934,10 +956,16 @@ gdip_pixel_stream_set_next (StreamingState *state, unsigned int pixel_value)
 		 * CAIRO_FORMAT_ARGB. This makes this section very easy to implement. If native
 		 * support for 15- and 16-bit pixel formats needs to be added in the future, though,
 		 * then this is where it needs to be done.
+		 *
+		 * In order to simplify advancing the state->scan pointer, the state->pixels_per_byte
+		 * member is set to the number of bytes per pixel, negated. That is, for 24-bit
+		 * formats, it is set to -3, and for 32-bit formats, it is set to -4.
+		 *
+		 * Note that pixel streams do not support 48- and 64-bit data at this time.
 		 */
 		*(unsigned int *)state->scan = pixel_value;
 
-		state->scan += 4;
+		state->scan -= state->pixels_per_byte;
 		state->x++;
 
 		if (state->x >= (state->region.X + state->region.Width)) {
@@ -946,7 +974,7 @@ gdip_pixel_stream_set_next (StreamingState *state, unsigned int pixel_value)
 
 			state->scan = (unsigned char *)(state->data->Scan0)
 			            + state->y * state->data->Stride
-			            + state->x * 4;
+			            + state->x * -state->pixels_per_byte;
 		}
 	}
 }
@@ -1289,12 +1317,12 @@ GdipBitmapUnlockBits (GpBitmap *bitmap, GdipBitmapData *locked_data)
 		/* We need to copy the locked data back to the root data's Scan0
 		 * (but of course only if the ImageLockMode specified writability)
 		 */
-		if ((locked_data->Reserved & GBD_READ_ONLY) != 0) {
+		if ((locked_data->Reserved & GBD_READ_ONLY) == 0) {
 			/* FIXME: destRect isn't necessarily equal to srcRect. this is a bug (the old implementation had it too). */
 			Rect srcRect = { 0, 0, locked_data->Width, locked_data->Height };
 			Rect destRect = srcRect;
 
-			GpStatus status = gdip_bitmap_change_rect_pixel_format (root_data, &srcRect, locked_data, &destRect);
+			GpStatus status = gdip_bitmap_change_rect_pixel_format (locked_data, &srcRect, root_data, &destRect);
 			if (status != Ok)
 				return status;
 		}
