@@ -1,6 +1,6 @@
 /*
  * Copyright (c) 2004 Ximian
- * Copyright (c) 2004-2005 Novell, Inc.
+ * Copyright (c) 2004-2006 Novell, Inc.
  * 
  * Permission is hereby granted, free of charge, to any person obtaining a copy of this software 
  * and associated documentation files (the "Software"), to deal in the Software without restriction, 
@@ -18,7 +18,8 @@
  * OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
  * 
  * Author:
- *          Jordi Mas i Hernandez <jordi@ximian.com>, 2004
+ *          Jordi Mas i Hernandez <jordi@ximian.com>, 2004-2006
+ *          Peter Dennis Bartok <pbartok@novell.com>
  */
 #include <stdio.h>
 #include "gdip.h"
@@ -111,9 +112,13 @@ GdipDeletePrivateFontCollection (GpFontCollection **font_collection)
 		return InvalidParameter;
 
 	if (*font_collection) {
-		FcFontSetDestroy ((*font_collection)->fontset);
-		FcConfigDestroy ((*font_collection)->config);
-		GdipFree ((void *)font_collection);
+		if ((*font_collection)->fontset != NULL) {
+			FcFontSetDestroy ((*font_collection)->fontset);
+		}
+		if ((*font_collection)->config != NULL) {
+			FcConfigDestroy ((*font_collection)->config);
+		}
+		GdipFree ((void *)*font_collection);
 	}
 
 	return Ok;
@@ -163,7 +168,9 @@ GdipDeleteFontFamily (GpFontFamily *fontFamily)
 	}	
 	
 	if (delete) {
-		FcPatternDestroy (fontFamily->pattern);
+		if (fontFamily->allocated) {
+			FcPatternDestroy (fontFamily->pattern);
+		}
 		GdipFree (fontFamily);
 	}
 	
@@ -204,10 +211,8 @@ GdipGetFontCollectionFamilyCount (GpFontCollection *font_collection, int *numFou
 }
 
 GpStatus
-GdipGetFontCollectionFamilyList (GpFontCollection *font_collection, int num_sought, GpFontFamily **gpfamilies, int *num_found)
+GdipGetFontCollectionFamilyList (GpFontCollection *font_collection, int num_sought, GpFontFamily *gpfamilies[], int *num_found)
 {
-	GpFontFamily **gpfam = gpfamilies;
-	FcPattern **pattern =  font_collection->fontset->fonts;
 	int i;
 
 	if (!font_collection || !gpfamilies || !num_found)
@@ -216,12 +221,12 @@ GdipGetFontCollectionFamilyList (GpFontCollection *font_collection, int num_soug
 	if (font_collection->config)
 		gdip_createPrivateFontSet (font_collection);
 
-	for (i = 0; i < font_collection->fontset->nfont; gpfam++, pattern++, i++) {
-		gdip_createFontFamily (gpfam);
-		(*gpfam)->pattern = *pattern;
-		(*gpfam)->allocated = FALSE;
+	for (i = 0; i < font_collection->fontset->nfont; i++) {
+		gdip_createFontFamily(&gpfamilies[i]);
+		gpfamilies[i]->pattern = font_collection->fontset->fonts[i];
+		gpfamilies[i]->allocated = FALSE;
 	}
-
+	
 	*num_found = font_collection->fontset->nfont;
 	return Ok;  
 }
@@ -305,7 +310,7 @@ GdipGetGenericFontFamilySansSerif (GpFontFamily **nativeFamily)
 {
 	const WCHAR MSSansSerif[] = {'M','S',' ','S','a','n','s',' ', 'S','e','r','i','f', 0};
 	
-	if (!familySansSerif) 
+	if (ref_familySansSerif == 0) 
 		GdipCreateFontFamilyFromName (MSSansSerif, NULL, &familySansSerif);    
 	
 	ref_familySansSerif++;
@@ -318,7 +323,7 @@ GdipGetGenericFontFamilySerif (GpFontFamily **nativeFamily)
 {
 	const WCHAR Serif[] = {'S','e','r','i','f', 0};
 	
-	if (!familySerif)
+	if (ref_familySerif == 0)
 		GdipCreateFontFamilyFromName (Serif, NULL, &familySerif);
 	
 	ref_familySerif++;	
@@ -331,7 +336,7 @@ GdipGetGenericFontFamilyMonospace (GpFontFamily **nativeFamily)
 {
 	const WCHAR Serif[] = {'S','e','r','i','f', 0};
 	
-	if (!familyMonospace)
+	if (ref_familyMonospace == 0)
 		GdipCreateFontFamilyFromName (Serif, NULL, &familyMonospace);    
 		
 	ref_familyMonospace++;
@@ -340,36 +345,32 @@ GdipGetGenericFontFamilyMonospace (GpFontFamily **nativeFamily)
 }
 
 FT_Face
-gdip_cairo_ft_font_lock_face (cairo_font_face_t *cairofnt)
+gdip_cairo_ft_font_lock_face (cairo_font_face_t *cairofnt, cairo_scaled_font_t **scaled_ft)
 {
-	cairo_scaled_font_t *scaled_ft;
+	FT_Face face;
 	cairo_matrix_t matrix1, matrix2;
-        cairo_font_options_t *options;
+        cairo_font_options_t *options = NULL;
 	
 	options = cairo_font_options_create ();
 		
-	scaled_ft = cairo_scaled_font_create (cairofnt,
+	cairo_matrix_init (&matrix1, 1, 0, 0, 1, 0, 0);
+	cairo_matrix_init (&matrix2, 1, 0, 0, 1, 0, 0);
+	*scaled_ft = cairo_scaled_font_create (cairofnt,
 					      &matrix1,
 					      &matrix2,
 					      options);
-	return cairo_ft_scaled_font_lock_face (scaled_ft);
+
+	face = cairo_ft_scaled_font_lock_face (*scaled_ft);
+
+	cairo_font_options_destroy (options);
+	return face;
 }
 
 void
-gdip_cairo_ft_font_unlock_face (cairo_font_face_t *cairofnt)
+gdip_cairo_ft_font_unlock_face (cairo_scaled_font_t* scaled_ft)
 {
-	cairo_scaled_font_t *scaled_ft;
-	cairo_matrix_t matrix1, matrix2;
-	cairo_font_options_t *options = NULL;
-
-	options = cairo_font_options_create ();	
-	
-	scaled_ft = cairo_scaled_font_create (cairofnt,
-					      &matrix1,
-					      &matrix2,
-					      options);
-	
 	cairo_ft_scaled_font_unlock_face (scaled_ft);
+	cairo_scaled_font_destroy (scaled_ft);
 }
 
 GpStatus
@@ -389,10 +390,11 @@ GdipGetEmHeight (GDIPCONST GpFontFamily *family, GpFontStyle style, short *EmHei
 	GdipCreateFont (family, 0.0f, style, UnitPoint, &font);
 
 	if (font) {
+		cairo_scaled_font_t* scaled_ft;
 		FT_Face	face;
 		TT_VertHeader *pVert;
 
-		face = gdip_cairo_ft_font_lock_face(font->cairofnt);
+		face = gdip_cairo_ft_font_lock_face (font->cairofnt, &scaled_ft);
 
 		pVert = FT_Get_Sfnt_Table(face, ft_sfnt_vhea);
 		if (pVert) {
@@ -402,7 +404,7 @@ GdipGetEmHeight (GDIPCONST GpFontFamily *family, GpFontStyle style, short *EmHei
 		} else {
 			rslt = 0;
 		}
-		gdip_cairo_ft_font_unlock_face(font->cairofnt);
+		gdip_cairo_ft_font_unlock_face(scaled_ft);
 		GdipDeleteFont (font);
 	}
 
@@ -428,17 +430,18 @@ GdipGetCellAscent (GDIPCONST GpFontFamily *family, GpFontStyle style, short *Cel
 	GdipCreateFont (family, 0.0f, style, UnitPoint, &font);
 
 	if (font){
+		cairo_scaled_font_t* scaled_ft;
                 FT_Face face;
 		TT_HoriHeader *pHori;
 
-                face = gdip_cairo_ft_font_lock_face(font->cairofnt);
+		face = gdip_cairo_ft_font_lock_face (font->cairofnt, &scaled_ft);
 
 		pHori = FT_Get_Sfnt_Table (face, ft_sfnt_hhea);
 
 		if (pHori)
 			rslt = pHori->Ascender;
 
-		gdip_cairo_ft_font_lock_face(font->cairofnt);
+		gdip_cairo_ft_font_unlock_face(scaled_ft);
 		GdipDeleteFont (font);
 	}
 
@@ -464,17 +467,18 @@ GdipGetCellDescent (GDIPCONST GpFontFamily *family, GpFontStyle style, short *Ce
 	GdipCreateFont (family, 0.0f, style, UnitPoint, &font);
 
 	if (font){
+		cairo_scaled_font_t* scaled_ft;
                 FT_Face face;
 		TT_HoriHeader* pHori;
 
-                face = gdip_cairo_ft_font_lock_face(font->cairofnt);
+		face = gdip_cairo_ft_font_lock_face (font->cairofnt, &scaled_ft);
 
 		pHori = FT_Get_Sfnt_Table (face, ft_sfnt_hhea);
 
 		if (pHori)
 			rslt = -pHori->Descender;
 
-		gdip_cairo_ft_font_unlock_face(font->cairofnt);
+		gdip_cairo_ft_font_unlock_face(scaled_ft);
 		GdipDeleteFont (font);
 	}
 
@@ -500,10 +504,11 @@ GdipGetLineSpacing (GDIPCONST GpFontFamily *family, GpFontStyle style, short *Li
 	GdipCreateFont (family, 0.0f, style, UnitPoint, &font);
 
 	if (font){
+		cairo_scaled_font_t* scaled_ft;
                 FT_Face face;
 		TT_HoriHeader *pHori;
 
-                face = gdip_cairo_ft_font_lock_face(font->cairofnt);
+		face = gdip_cairo_ft_font_lock_face (font->cairofnt, &scaled_ft);
 
 		pHori = FT_Get_Sfnt_Table (face, ft_sfnt_hhea);
 		if (pHori) {
@@ -514,7 +519,7 @@ GdipGetLineSpacing (GDIPCONST GpFontFamily *family, GpFontStyle style, short *Li
 			rslt = 0;
 		}
 
-		gdip_cairo_ft_font_unlock_face(font->cairofnt);
+		gdip_cairo_ft_font_unlock_face(scaled_ft);
 		GdipDeleteFont (font);
 	}
 
@@ -551,7 +556,7 @@ GdipCreateFont (GDIPCONST GpFontFamily* family, float emSize, GpFontStyle style,
 
 	r = FcPatternGetString (family->pattern, FC_FAMILY, 0, &str);
 	
-	gdip_unitConversion (unit, UnitPixel, emSize, &sizeInPixels);
+	gdip_unit_conversion (unit, UnitPixel, gdip_get_display_dpi(), gtMemoryBitmap, emSize, &sizeInPixels);
 		
 	result = (GpFont *) GdipAlloc (sizeof (GpFont));
 	result->sizeInPixels = sizeInPixels;
@@ -568,11 +573,25 @@ GdipCreateFont (GDIPCONST GpFontFamily* family, float emSize, GpFontStyle style,
 
 	cairofnt = _cairo_toy_font_face_create ((const char*) str, slant, weight);
 
-	if (cairofnt == NULL)
-		return GenericError;	
+	if (cairofnt == NULL) {
+		GdipFree(result);
+		return GenericError;
+	}
+
+	result->face = GdipAlloc(strlen((char *)str) + 1);
+	if (!result->face) {
+		GdipFree(result);
+		return OutOfMemory;
+	}
+
+	memcpy(result->face, str, strlen((char *)str) + 1);
 
 	result->cairofnt = cairofnt;
         result->style = style;
+	result->emSize = emSize;
+	result->unit = unit;
+	result->family = (GpFontFamily*) family;
+	result->style = style;
 	*font = result;	        		
 	return Ok;
 }
@@ -581,7 +600,7 @@ GpStatus
 GdipGetHfont(GpFont* font, void **Hfont)
 {
 	if (font) {
-		*Hfont=font->wineHfont;
+		*Hfont=font;
 		return(Ok);
 	}
 	return InvalidParameter;
@@ -595,6 +614,7 @@ GdipDeleteFont (GpFont* font)
 		return InvalidParameter;
 
 	cairo_font_face_destroy (font->cairofnt);
+	GdipFree ((void *)font->face);
 	GdipFree ((void *)font);
 	return Ok;	       
 }
@@ -605,20 +625,224 @@ GdipCreateFontFromDC(void *hdc, GpFont **font)
 	return(NotImplemented);
 }
 
-GpStatus
-GdipCreateFontFromHfont(void *hfont, GpFont **font, void *lf)
+static void
+gdip_logfont_from_font(GpFont *font, GpGraphics *graphics, void *lf, bool ucs2)
 {
-	return(NotImplemented);
+	LOGFONTA		*logFont;
+
+	logFont = (LOGFONTA *)lf;
+
+	logFont->lfHeight = -(font->sizeInPixels);
+	logFont->lfWidth = 0;
+	logFont->lfEscapement = 0;	// FIXME
+	logFont->lfOrientation = logFont->lfEscapement;
+	if (font->style & FontStyleBold) {
+		logFont->lfWeight = 700;
+	} else {
+		logFont->lfWeight = 400;
+	}
+
+	if (font->style & FontStyleItalic) {
+		logFont->lfItalic = 1;
+	} else {
+		logFont->lfItalic = 0;
+	}
+
+	if (font->style & FontStyleUnderline) {
+		logFont->lfUnderline = 1;
+	} else {
+		logFont->lfUnderline = 0;
+	}
+
+	if (font->style & FontStyleStrikeout) {
+		logFont->lfStrikeOut = 1;
+	} else {
+		logFont->lfStrikeOut = 0;
+	}
+
+	logFont->lfCharSet = 1;	// DEFAULT_CHARSET
+	logFont->lfOutPrecision = 0;
+	logFont->lfClipPrecision = 0;
+	if (graphics != NULL) {
+		switch(graphics->text_mode) {
+			case TextRenderingHintSystemDefault: {
+				logFont->lfQuality = 0;
+				break;
+			}
+	
+			case TextRenderingHintSingleBitPerPixelGridFit:
+			case TextRenderingHintSingleBitPerPixel:
+			case TextRenderingHintAntiAliasGridFit:
+			case TextRenderingHintAntiAlias: {
+				logFont->lfQuality = 3;	// ANTIALIASED_QUALITY;
+				break;
+			}
+	
+			case TextRenderingHintClearTypeGridFit: {
+				logFont->lfQuality = 5;	// CLEARTYPE_QUALITY
+				break;
+			}
+		}
+	} else {
+		logFont->lfQuality = 0;
+	}
+
+	logFont->lfPitchAndFamily = 0;
+	if (ucs2) {
+		utf8_to_ucs2((const gchar *)font->face, (gunichar2 *)logFont->lfFaceName, LF_FACESIZE);
+	} else {
+		memcpy(logFont->lfFaceName, font->face, LF_FACESIZE);
+		logFont->lfFaceName[LF_FACESIZE - 1] = '\0';
+	}
+}
+
+GpStatus
+GdipCreateFontFromHfontA(void *hfont, GpFont **font, void *lf)
+{
+	GpFont			*src_font;
+	GpFont			*result;
+	cairo_font_face_t	*face;
+	cairo_font_slant_t	slant;
+	cairo_font_weight_t	weight;
+
+	src_font = (GpFont *)hfont;
+
+	result = (GpFont *) GdipAlloc (sizeof (GpFont));
+	result->sizeInPixels = src_font->sizeInPixels;
+	result->style = src_font->style;
+	result->family = src_font->family;
+	result->style = src_font->style;
+	result->emSize = src_font->emSize;
+	result->unit = src_font->unit;
+
+        if ((result->style & FontStyleBold) == FontStyleBold) {
+                weight = CAIRO_FONT_WEIGHT_BOLD;
+	} else {
+		weight = CAIRO_FONT_WEIGHT_NORMAL;
+	}
+
+        if ((result->style & FontStyleItalic) == FontStyleItalic) {
+                slant = CAIRO_FONT_SLANT_ITALIC;
+	} else {
+		slant = CAIRO_FONT_SLANT_NORMAL;
+	}
+	result->face = GdipAlloc(strlen((char *)src_font->face) + 1);
+	if (!result->face) {
+		GdipFree(result);
+		return OutOfMemory;
+	}
+
+	memcpy(result->face, src_font->face, strlen((char *)src_font->face) + 1);
+
+	result->cairofnt = _cairo_toy_font_face_create ((const char*) src_font->face, slant, weight);
+
+	if (result->cairofnt == NULL) {
+		GdipFree(result);
+		return GenericError;
+	}
+
+	*font = result;
+
+	gdip_logfont_from_font(result, NULL, lf, FALSE);
+
+	return Ok;
+}
+
+GpStatus
+GdipGetLogFontW(GpFont *font, GpGraphics *graphics, void *lf)
+{
+	if (!font || !lf) {
+		return(InvalidParameter);
+	}
+	gdip_logfont_from_font(font, graphics, lf, TRUE);
+	return Ok;
 }
 
 GpStatus
 GdipGetLogFontA(GpFont *font, GpGraphics *graphics, void *lf)
 {
+		
 	if (!font || !lf) {
 		return(InvalidParameter);
 	}
 	
-	return NotImplemented;
+	gdip_logfont_from_font(font, graphics, lf, FALSE);
+	return Ok;
+}
+
+GpStatus
+gdip_create_font_from_logfont(void *hdc, void *lf, GpFont **font, bool ucs2)
+{
+	GpFont			*result;
+	cairo_font_face_t	*face;
+	cairo_font_slant_t	slant;
+	cairo_font_weight_t	weight;
+	LOGFONTA		*logfont;
+
+	logfont = (LOGFONTA *)lf;
+
+	result = (GpFont *) GdipAlloc(sizeof(GpFont));
+	if (logfont->lfHeight < 0) {
+		result->sizeInPixels = -(logfont->lfHeight);
+	} else {
+		result->sizeInPixels = logfont->lfHeight;	// Fixme - convert units
+	}
+	result->style = 0;
+	result->family = 0;
+	/* Fixme - this is wrong, but I don't know of a quick way to get the emSize */
+	result->emSize = result->sizeInPixels;
+	result->unit = UnitPixel;
+
+	if (logfont->lfItalic) {
+		result->style |= FontStyleItalic;
+		slant = CAIRO_FONT_SLANT_ITALIC;
+	} else {
+		slant = CAIRO_FONT_SLANT_NORMAL;
+	}
+
+	if (logfont->lfWeight > 400) {
+		result->style |= FontStyleBold;
+		weight = CAIRO_FONT_WEIGHT_BOLD;
+	} else {
+		weight = CAIRO_FONT_WEIGHT_NORMAL;
+	}
+
+	if (logfont->lfUnderline) {
+		result->style |= FontStyleUnderline;
+	}
+	if (logfont->lfStrikeOut) {
+		result->style |= FontStyleStrikeout;
+	}
+
+	if (ucs2) {
+		result->face = (unsigned char*) ucs2_to_utf8 ((const gunichar2 *)logfont->lfFaceName, -1);
+	} else {
+		result->face = GdipAlloc(LF_FACESIZE);
+		memcpy(result->face, logfont->lfFaceName, LF_FACESIZE);
+		result->face[LF_FACESIZE - 1] = '\0';
+	}
+
+	result->cairofnt = _cairo_toy_font_face_create ((const char *)result->face, slant, weight);
+	if (result->cairofnt == NULL) {
+		GdipFree(result);
+		return GenericError;
+	}
+
+	*font = result;
+
+	return Ok;
+}
+
+GpStatus
+GdipCreateFontFromLogfontA(void *hdc, GDIPCONST LOGFONTA *logfont, GpFont **font)
+{
+	return gdip_create_font_from_logfont(hdc, (void *)logfont, font, FALSE);
+}
+
+GpStatus
+GdipCreateFontFromLogfontW(void *hdc, GDIPCONST LOGFONTW *logfont, GpFont **font)
+{
+	return gdip_create_font_from_logfont(hdc, (void *)logfont, font, TRUE);
 }
 
 GpStatus
@@ -669,3 +893,50 @@ GetFontMetrics(GpGraphics *graphics, GpFont *font, int *ascent, int *descent)
 
 	return TRUE;
 }
+
+
+GpStatus
+GdipGetFontHeight (GDIPCONST GpFont *font, GDIPCONST GpGraphics *graphics, float *height)
+{
+	short emHeight, lineSpacing;
+	float emSize;
+
+	if (!font || !height || !graphics)
+		return InvalidParameter;
+
+	/* Operations in display dpi's */	
+	gdip_unit_conversion (font->unit, UnitPixel, gdip_get_display_dpi (), gtMemoryBitmap, font->emSize, &emSize);
+	GdipGetEmHeight (font->family, font->style, &emHeight);
+	GdipGetLineSpacing (font->family, font->style, &lineSpacing);
+
+	*height = lineSpacing * (emSize / emHeight);
+	gdip_unit_conversion (UnitPixel, graphics->page_unit, gdip_get_display_dpi (), graphics->type, *height, height);
+	return Ok;
+}
+
+GpStatus
+GdipGetFontHeightGivenDPI (GDIPCONST GpFont *font, float dpi, float *height)
+{
+	short emHeight, lineSpacing;
+
+	if (!font || !height)
+		return InvalidParameter;
+
+	GdipGetEmHeight (font->family, font->style, &emHeight);
+	GdipGetLineSpacing (font->family, font->style, &lineSpacing);
+	*height = lineSpacing * (font->emSize / emHeight);
+	gdip_unit_conversion (font->unit, UnitInch, dpi, gtMemoryBitmap, *height, height);
+	*height = *height * dpi;
+	return Ok;
+}
+
+GpStatus
+GdipGetFontSize (GpFont *font, float *size)
+{
+	if (!font ||!size)
+		return InvalidParameter;
+
+	*size = font->emSize;
+	return Ok;
+}
+
