@@ -11,7 +11,7 @@
  *	Geoff Norton (gnorton@customerdna.com)
  *      Jonathan Gilbert (logic@deltaq.org)
  *
- * Copyright (C) Novell, Inc. 2003-2005. http://www.novell.com
+ * Copyright (C) 2003-2006 Novell, Inc (http://www.novell.com)
  */
 
 #ifndef _GDIP_H
@@ -208,7 +208,8 @@ typedef enum {
     UnitPoint	 = 3,
     UnitInch	 = 4,
     UnitDocument = 5,
-    UnitMillimeter = 6
+    UnitMillimeter = 6,
+    UnitCairoPoint = 7
 } GpUnit, Unit;
 
 typedef enum {
@@ -347,7 +348,8 @@ typedef enum {
 	gtUndefined,
 	gtX11Drawable,
 	gtMemoryBitmap,
-	gtOSXDrawable
+	gtOSXDrawable,
+	gtPostScript
 } GraphicsType;
 
 typedef enum{
@@ -666,11 +668,6 @@ typedef struct {
         float X, Y;
 } GpPointF, PointF;
 
-typedef struct {
-        GpRectF*        rects;
-        int             cnt;
-} GpRegion;
-
 typedef struct {
 	int first, length;
 } CharacterRange;
@@ -702,6 +699,52 @@ typedef struct {
         BOOL changed; /* flag to mark if pen is changed and needs setup */
 } GpPen;
 
+typedef struct {
+	GpFillMode fill_mode;
+	int count;
+	GByteArray *types;
+	GArray *points;
+	BOOL start_new_fig; /* Flag to keep track if we need to start a new figure */
+} GpPath;
+
+typedef struct {
+        int Count;
+        PointF *Points;
+        byte *Types;
+} GpPathData;
+
+typedef struct {
+	GpPath *path;
+	int markerPosition; /* The start position of next marker, index of (marker type) + 1  */
+	int subpathPosition; /* The start position of next subpath, index of (start type) */
+	int pathTypePosition; /* The position to get the next path type inside a subpath */
+} GpPathIterator;
+
+/* internal (private) structure */
+typedef struct {
+	int X;
+	int Y;
+	int Width;
+	int Height;
+	unsigned char *Mask;
+	BOOL reduced;
+} GpRegionBitmap;
+
+/* internal (private) structure */
+typedef struct GpPathTree {
+	CombineMode	mode;
+	GpPath*		path;
+	struct GpPathTree*	branch1;
+	struct GpPathTree*	branch2;
+} GpPathTree;
+
+typedef struct {
+	guint32		type;
+        int		cnt;
+        GpRectF*	rects;
+	GpPathTree*	tree;
+	GpRegionBitmap*	bitmap;
+} GpRegion;
  
 typedef struct {
 	cairo_matrix_t		matrix;
@@ -743,6 +786,8 @@ typedef struct {
 	float		aa_offset_y;
 	int		render_origin_x;
 	int		render_origin_y;
+	float		dpi_x;
+	float		dpi_y;
 } GpGraphics;
 
 
@@ -805,32 +850,8 @@ typedef struct {
 	GpImage		image;
         int             cairo_format;
 	BitmapData	data;
-	void		*hBitmapDC;
-	void		*hInitialBitmap;
-	void		*hBitmap;	
 } GpBitmap;
 
-typedef struct {
-	GpFillMode fill_mode;
-	int count;
-	GByteArray *types;
-	GArray *points;
-	BOOL start_new_fig; /* Flag to keep track if we need to start a new figure */
-} GpPath;
-
-typedef struct {
-        int Count;
-        PointF *Points;
-        byte *Types;
-} GpPathData;
-
-typedef struct {
-	GpPath *path;
-	int markerPosition; /* The start position of next marker, index of (marker type) + 1  */
-	int subpathPosition; /* The start position of next subpath, index of (start type) */
-	int pathTypePosition; /* The position to get the next path type inside a subpath */
-} GpPathIterator;
-
 typedef struct {
         FcFontSet*  fontset;
         /* Only for private collections */
@@ -852,9 +873,46 @@ typedef struct {
         cairo_font_face_t*  cairofnt;
         float               sizeInPixels;
         GpFontStyle         style;
-        void                *wineHfont;
+        unsigned char       *face;
+	GpFontFamily	    *family;
+	float               emSize;
+	GpUnit              unit;
 } GpFont;
 
+
+typedef struct {
+    LONG lfHeight;
+    LONG lfWidth;
+    LONG lfEscapement;
+    LONG lfOrientation;
+    LONG lfWeight;
+    BYTE lfItalic;
+    BYTE lfUnderline;
+    BYTE lfStrikeOut;
+    BYTE lfCharSet;
+    BYTE lfOutPrecision;
+    BYTE lfClipPrecision;
+    BYTE lfQuality;
+    BYTE lfPitchAndFamily;
+    WCHAR lfFaceName[LF_FACESIZE];
+} LOGFONTW;
+
+typedef struct {
+    LONG lfHeight;
+    LONG lfWidth;
+    LONG lfEscapement;
+    LONG lfOrientation;
+    LONG lfWeight;
+    BYTE lfItalic;
+    BYTE lfUnderline;
+    BYTE lfStrikeOut;
+    BYTE lfCharSet;
+    BYTE lfOutPrecision;
+    BYTE lfClipPrecision;
+    BYTE lfQuality;
+    BYTE lfPitchAndFamily;
+    char lfFaceName[LF_FACESIZE];
+} LOGFONTA;
  
 typedef struct
 {
@@ -935,7 +993,8 @@ typedef struct
 	int x, y;               /* the offset of the next byte that will be loaded, once the buffer is depleted */
 	unsigned short buffer;
 	int p;                  /* index of pixel within 'buffer' that was returned by the last call to gdip_pixel_stream_get_next () */
-	int one_pixel_mask, one_pixel_shift, pixels_per_byte;
+	int one_pixel_mask, one_pixel_shift;
+	int pixels_per_byte;	/* a negative value is used to indicate a count of bytes per pixel for depths of more than 8 bits */
 
 	BitmapData *data;
 	unsigned char *scan;
@@ -955,7 +1014,7 @@ void gdip_image_init              (GpImage *image);
 void gdip_bitmap_init  (GpBitmap *bitmap);
 GpBitmap *gdip_bitmap_new   (void);
 void gdip_bitmap_dispose (GpBitmap *bitmap);
-void gdip_bitmap_clone (GpBitmap *bitmap, GpBitmap **clonedbitmap);
+GpStatus gdip_bitmap_clone (GpBitmap *bitmap, GpBitmap **clonedbitmap);
 
 void gdip_graphics_init (GpGraphics *graphics, cairo_surface_t *surface);
 GpGraphics *gdip_graphics_new (cairo_surface_t *surface);
@@ -1069,6 +1128,8 @@ GpStatus GdipGetTextContrast(GpGraphics *graphics, UINT *contrast);
 GpStatus GdipSetSmoothingMode(GpGraphics *graphics, SmoothingMode mode);
 GpStatus GdipGetSmoothingMode(GpGraphics *graphics, SmoothingMode *mode);
 GpStatus GdipSetVisibleClip_linux (GpGraphics *graphics, GpRect *rect);
+GpStatus GdipTranslateClip (GpGraphics *graphics, float dx, float dy);
+GpStatus GdipTranslateClipI (GpGraphics *graphics, UINT dx, UINT dy);
 
 /* Status */
 GpStatus gdip_get_status (cairo_status_t status);
@@ -1249,6 +1310,7 @@ GpStatus GdipCreateRegion(GpRegion **region);
 GpStatus GdipCreateRegionRect(GDIPCONST GpRectF *rect, GpRegion **region);
 GpStatus GdipCreateRegionRectI(GDIPCONST GpRect *rect, GpRegion **region);
 GpStatus GdipCreateRegionPath(GpPath *path, GpRegion **region);
+GpStatus GdipCreateRegionRgnData (GDIPCONST BYTE *regionData, int size, GpRegion **region);
 GpStatus GdipCloneRegion(GpRegion *region, GpRegion **cloneRegion);
 GpStatus GdipDeleteRegion(GpRegion *region);
 GpStatus GdipSetInfinite(GpRegion *region);
@@ -1299,7 +1361,10 @@ void GdipFree (void *ptr);
 /* Utility*/
 int fcmp (double x1, double x2, double epsilon);
 float gdip_get_display_dpi();
-void gdip_unitConversion(Unit fromUnit, Unit toUnit, float nSrc, float* nTrg);
+float gdip_unitx_convgr (GpGraphics *graphics, float nSrc);
+float gdip_unity_convgr (GpGraphics *graphics, float nSrc);
+void gdip_unit_conversion (Unit from, Unit to, float dpi, GraphicsType type, float nSrc, float* nTrg);
+
 cairo_content_t from_cairoformat_to_content (cairo_format_t format);
 
 void gdip_rect_expand_by (GpRectF *rect, GpPointF *point);
