@@ -2,6 +2,7 @@
  * graphics-path.c
  *
  * Copyright (C) 2003, Novell Inc.
+ * Copyright (C) 2006 Novell, Inc (http://www.novell.com)
  * 
  * Permission is hereby granted, free of charge, to any person obtaining a copy of this software 
  * and associated documentation files (the "Software"), to deal in the Software without restriction, 
@@ -83,6 +84,23 @@ int_to_float (const GpPoint *pts, int count)
         return p;
 }
 
+/* return TRUE if the specified path has (at least one) curves, FALSE otherwise */
+static BOOL
+gdip_path_has_curve (GpPath *path)
+{
+	int i;
+
+	if (!path)
+		return FALSE;
+
+	for (i = 0; i < path->count; i++) {
+		if (g_array_index (path->types, byte, i) == PathPointTypeBezier)
+			return TRUE;
+	}
+
+	return FALSE;
+}
+
 static void
 append (GpPath *path, float x, float y, GpPathPointType type)
 {
@@ -383,11 +401,12 @@ GdipClosePathFigure (GpPath *path)
 	byte current;
 	g_return_val_if_fail (path != NULL, InvalidParameter);
 
-	current = g_array_index (path->types, byte, path->count - 1);
-	g_byte_array_remove_index (path->types, path->count - 1);
-	current |= PathPointTypeCloseSubpath;
-	g_byte_array_append (path->types, &current, 1);
-
+	if (path->count > 0) {
+		current = g_array_index (path->types, byte, path->count - 1);
+		g_byte_array_remove_index (path->types, path->count - 1);
+		current |= PathPointTypeCloseSubpath;
+		g_byte_array_append (path->types, &current, 1);
+	}
 	path->start_new_fig = TRUE;
 
 	return Ok;
@@ -443,6 +462,9 @@ GdipSetPathMarker (GpPath *path)
 	byte current;
 	g_return_val_if_fail (path != NULL, InvalidParameter);
 
+	if (path->count == 0)
+		return Ok;
+
 	current = g_array_index (path->types, byte, path->count - 1);
 
         g_byte_array_remove_index (path->types, path->count - 1);
@@ -463,6 +485,10 @@ GdipClearPathMarkers (GpPath *path)
 
 	g_return_val_if_fail (path != NULL, InvalidParameter);
 
+	/* shortcut to avoid allocations */
+	if (path->count == 0)
+		return Ok;
+
 	cleared = g_byte_array_new ();
 
         for (i = 0; i < path->count; i++) {
@@ -492,6 +518,9 @@ GdipReversePath (GpPath *path)
 	g_return_val_if_fail (path != NULL, InvalidParameter);
 
 	length = path->count;
+	/* shortcut to avoid allocations */
+	if (length <= 1)
+		return Ok;
 
 	types = g_byte_array_sized_new (length);
 	points = g_array_sized_new (FALSE, TRUE, sizeof (GpPointF), length);
@@ -560,7 +589,19 @@ append_arc (GpPath *path, bool start, float x, float y, float width, float heigh
         /* angles in radians */        
         float alpha = startAngle * PI / 180;
         float beta = endAngle * PI / 180;
-        float delta = beta - alpha;
+
+        /* adjust angles for ellipses */
+	alpha = atan2 (rx * sin (alpha), ry * cos (alpha));
+	beta = atan2 (rx * sin (beta), ry * cos (beta));
+
+	if (abs (beta - alpha) > M_PI){
+		if (beta > alpha)
+			beta -= 2 * PI;
+		else
+			alpha -= 2 * PI;
+	}
+
+	float delta = beta - alpha;
         float bcp = 4.0 / 3 * (1 - cos (delta / 2)) / sin (delta / 2);
 
         double sin_alpha = sin (alpha);
@@ -590,10 +631,11 @@ append_arcs (GpPath *path, float x, float y, float width, float height, float st
 {
 	int i;
 	float drawn = 0;
-	float endAngle = startAngle + sweepAngle;	
+	float endAngle = startAngle + sweepAngle;
 	int sign = (endAngle > 0) ? 1 : -1;
-	int increment = sign * 90; 
-	
+	int increment = sign * 90;
+	bool enough = FALSE;
+
 	if (abs (sweepAngle) >= 360) {
 		GdipAddPathEllipse (path, x, y, width, height);
 		return;
@@ -604,20 +646,22 @@ append_arcs (GpPath *path, float x, float y, float width, float height, float st
 	for (i = 0; i < 4; i++) {
 		float current = startAngle + drawn;
 
-		/* we've drawn enough */
-		if (abs (current) >= abs (endAngle))
-			break;
-		
-		/* if we are not done yet */
-		float additional = abs (current + increment) < abs (endAngle) ?
-				increment : /* add the default increment */
-				endAngle - current; /* otherwise, add the remainder */
+		if (enough)
+			return;
 		
+		float additional;
+		if (abs (current + increment) < abs (endAngle))
+			additional = increment;  /* add the default increment */
+		else {
+			additional = endAngle - current; /* otherwise, add the remainder */
+			enough = TRUE;
+		}
+
 		append_arc (path,
 			    (i == 0) ? TRUE : FALSE,  /* only move to the starting pt in the 1st iteration */
 			    x, y, width, height,      /* bounding rectangle */
 			    current, current + additional);
-		drawn += additional;		
+		drawn += additional;
 	}
 }
 
@@ -653,6 +697,9 @@ GdipAddPathBeziers (GpPath *path, const GpPointF *points, int count)
 	GpPointF *tmp = (GpPointF *) points;
 	g_return_val_if_fail (path != NULL, InvalidParameter);
 	g_return_val_if_fail (points != NULL, InvalidParameter);
+	/* first bezier requires 4 points, other 3 more points */
+	if ((count < 4) || ((count % 3) != 1))
+		return InvalidParameter;
         
         append_point (path, *tmp, PathPointTypeLine);
         tmp++;
@@ -682,6 +729,8 @@ GdipAddPathCurve3 (GpPath *path, const GpPointF *points, int count,
         GpPointF *tangents;
 	g_return_val_if_fail (path != NULL, InvalidParameter);
 	g_return_val_if_fail (points != NULL, InvalidParameter);
+	if (count < 2)
+		return InvalidParameter;
 
         tangents = gdip_open_curve_tangents (CURVE_MIN_TERMS, points, count, tension);
 
@@ -703,6 +752,8 @@ GdipAddPathClosedCurve2 (GpPath *path, const GpPointF *points, int count, float
 {
         GpPointF *tangents;
 	g_return_val_if_fail (points != NULL, InvalidParameter);
+	if (count < 3)
+		return InvalidParameter;
 
         tangents = gdip_closed_curve_tangents (CURVE_MIN_TERMS, points, count, tension);
 
@@ -798,9 +849,13 @@ GdipAddPathPie (GpPath *path, float x, float y, float width, float height, float
         /* center */
         int cx = x + rx;
         int cy = y + ry;	
+
         /* angles in radians */        
         float alpha = startAngle * PI / 180;
 
+        /* adjust angle for ellipses */
+        alpha = atan2 (rx * sin (alpha), ry * cos (alpha));
+
         float sin_alpha = sin (alpha);
         float cos_alpha = cos (alpha);
 
@@ -829,6 +884,8 @@ GdipAddPathPolygon (GpPath *path, const GpPointF *points, int count)
         GpPointF *tmp = (GpPointF *) points;
 	g_return_val_if_fail (path != NULL, InvalidParameter);
 	g_return_val_if_fail (points != NULL, InvalidParameter);
+	if (count < 3)
+		return InvalidParameter;
         
         append_point (path, *tmp, PathPointTypeStart);
         tmp ++;
@@ -901,8 +958,9 @@ GpStatus
 GdipAddString (GpPath *path, const char *string, int length, 
                 const GpFontFamily *family, int style, float emSize,
                 const GpRectF *layoutRect, const GpStringFormat *format)
-{ 
-	return NotImplemented;
+{
+	g_warning ("GdipAddString isn't implemented in libgdiplus.");
+	return Ok; //NotImplemented;
 }
 
 /* MonoTODO */
@@ -911,7 +969,8 @@ GdipAddStringI (GpPath *path, const char *string, int length,
                 const GpFontFamily *family, int style, float emSize,
                 const GpRect *layoutRect, const GpStringFormat *format)
 {
-	return NotImplemented;
+	g_warning ("GdipAddStringI isn't implemented in libgdiplus.");
+	return Ok; //NotImplemented;
 }
 
 GpStatus
@@ -1067,11 +1126,205 @@ GdipAddPathPolygonI (GpPath *path, const GpPoint *points, int count)
         return s;
 }
 
-/* MonoTODO */
+/* nr_curve_flatten comes from Sodipodi's libnr (public domain) available from http://www.sodipodi.com/ */
+/* Mono changes: converted to float (from double), added recursion limit, use GArray */
+static BOOL
+nr_curve_flatten (float x0, float y0, float x1, float y1, float x2, float y2, float x3, float y3, float flatness, int level, GArray *points)
+{
+	float dx1_0, dy1_0, dx2_0, dy2_0, dx3_0, dy3_0, dx2_3, dy2_3, d3_0_2;
+	float s1_q, t1_q, s2_q, t2_q, v2_q;
+	float f2, f2_q;
+	float x00t, y00t, x0tt, y0tt, xttt, yttt, x1tt, y1tt, x11t, y11t;
+
+	dx1_0 = x1 - x0;
+	dy1_0 = y1 - y0;
+	dx2_0 = x2 - x0;
+	dy2_0 = y2 - y0;
+	dx3_0 = x3 - x0;
+	dy3_0 = y3 - y0;
+	dx2_3 = x3 - x2;
+	dy2_3 = y3 - y2;
+	f2 = flatness;
+	d3_0_2 = dx3_0 * dx3_0 + dy3_0 * dy3_0;
+	if (d3_0_2 < f2) {
+		float d1_0_2, d2_0_2;
+		d1_0_2 = dx1_0 * dx1_0 + dy1_0 * dy1_0;
+		d2_0_2 = dx2_0 * dx2_0 + dy2_0 * dy2_0;
+		if ((d1_0_2 < f2) && (d2_0_2 < f2)) {
+			goto nosubdivide;
+		} else {
+			goto subdivide;
+		}
+	}
+	f2_q = f2 * d3_0_2;
+	s1_q = dx1_0 * dx3_0 + dy1_0 * dy3_0;
+	t1_q = dy1_0 * dx3_0 - dx1_0 * dy3_0;
+	s2_q = dx2_0 * dx3_0 + dy2_0 * dy3_0;
+	t2_q = dy2_0 * dx3_0 - dx2_0 * dy3_0;
+	v2_q = dx2_3 * dx3_0 + dy2_3 * dy3_0;
+	if ((t1_q * t1_q) > f2_q) goto subdivide;
+	if ((t2_q * t2_q) > f2_q) goto subdivide;
+	if ((s1_q < 0.0) && ((s1_q * s1_q) > f2_q)) goto subdivide;
+	if ((v2_q < 0.0) && ((v2_q * v2_q) > f2_q)) goto subdivide;
+	if (s1_q >= s2_q) goto subdivide;
+
+ nosubdivide:
+	{
+	GpPointF pt;
+
+	pt.X = x3;
+	pt.Y = y3;
+	g_array_append_val (points, pt);
+	return TRUE;
+	}
+ subdivide:
+	/* things gets *VERY* memory intensive without a limit */
+	if (level >= FLATTEN_RECURSION_LIMIT)
+		return FALSE;
+
+	x00t = (x0 + x1) * 0.5;
+	y00t = (y0 + y1) * 0.5;
+	x0tt = (x0 + 2 * x1 + x2) * 0.25;
+	y0tt = (y0 + 2 * y1 + y2) * 0.25;
+	x1tt = (x1 + 2 * x2 + x3) * 0.25;
+	y1tt = (y1 + 2 * y2 + y3) * 0.25;
+	x11t = (x2 + x3) * 0.5;
+	y11t = (y2 + y3) * 0.5;
+	xttt = (x0tt + x1tt) * 0.5;
+	yttt = (y0tt + y1tt) * 0.5;
+
+	if (!nr_curve_flatten (x0, y0, x00t, y00t, x0tt, y0tt, xttt, yttt, flatness, level+1, points)) return FALSE;
+	if (!nr_curve_flatten (xttt, yttt, x1tt, y1tt, x11t, y11t, x3, y3, flatness, level+1, points)) return FALSE;
+	return TRUE;
+}
+
+static BOOL
+gdip_convert_bezier_to_lines (GpPath *path, int index, float flatness, GArray *flat_points, GByteArray *flat_types)
+{
+	GArray *points;
+	GpPointF start, first, second, end;
+	GpPointF pt;
+	byte type;
+	int i;
+
+	if ((index <= 0) || (index + 2 >= path->count))
+		return FALSE; /* bad path data */
+
+	start = g_array_index (path->points, GpPointF, index - 1);
+	first = g_array_index (path->points, GpPointF, index);
+	second = g_array_index (path->points, GpPointF, index + 1);
+	end = g_array_index (path->points, GpPointF, index + 2);
+
+	/* we can't add points directly to the original list as we could end up with too much recursion */
+	points = g_array_new (FALSE, FALSE, sizeof (GpPointF));
+	if (!nr_curve_flatten (start.X, start.Y, first.X, first.Y, second.X, second.Y, end.X, end.Y, flatness, 0, points)) {
+		/* curved path is too complex (i.e. would result in too many points) to render as a polygon */
+		g_array_free (points, TRUE);
+		return FALSE;
+	}
+
+	/* recursion was within limits, append the result to the original supplied list */
+	if (points->len > 0) {
+		g_array_append_val (flat_points, g_array_index (points, GpPointF, 0));
+		/* special case: first point type could be PathPointTypeStart */
+		type = (index == 0) ? PathPointTypeStart : PathPointTypeLine;
+		g_byte_array_append (flat_types, &type, 1);
+	}
+
+	/* always PathPointTypeLine */
+	type = PathPointTypeLine;
+	for (i = 1; i < points->len; i++) {
+		pt = g_array_index (points, GpPointF, i);
+		g_array_append_val (flat_points, pt);
+		g_byte_array_append (flat_types, &type, 1);
+	}
+
+	g_array_free (points, TRUE);
+	return TRUE;
+}
+
 GpStatus 
 GdipFlattenPath (GpPath *path, GpMatrix *matrix, float flatness)
 {
-	return NotImplemented;
+	GpStatus status = Ok;
+	GArray *points;
+	GByteArray *types;
+	int i;
+
+	g_return_val_if_fail (path != NULL, InvalidParameter);
+
+	/* apply matrix before flattening (as there's less points at this stage) */
+	if (matrix) {
+		status = GdipTransformPath (path, matrix);
+		if (status != Ok)
+			return status;
+	}
+
+	/* if no bezier are present then the path doesn't need to be flattened */
+	if (!gdip_path_has_curve (path))
+		return status;
+
+	points = g_array_new (FALSE, FALSE, sizeof (GpPointF));
+	types = g_byte_array_new ();
+
+	/* Iterate the current path and replace each bezier with multiple lines */
+	for (i = 0; i < path->count; i++) {
+		GpPointF point = g_array_index (path->points, GpPointF, i);
+		byte type = g_array_index (path->types, byte, i);
+
+		/* PathPointTypeBezier3 has the same value as PathPointTypeBezier */
+		if ((type & PathPointTypeBezier) == PathPointTypeBezier) {
+			if (!gdip_convert_bezier_to_lines (path, i, fabs (flatness), points, types)) {
+				/* uho, too much recursion - do not pass go, do not collect 200$ */
+				GpPointF pt;
+
+				/* free the the partial flat */
+				g_array_free (points, TRUE);
+				g_byte_array_free (types, TRUE);
+
+				/* mimic MS behaviour when recursion becomes a problem */
+				/* note: it's not really an empty rectangle as the last point isn't closing */
+				points = g_array_new (FALSE, FALSE, sizeof (GpPointF));
+				types = g_byte_array_new ();
+
+				pt.X = pt.Y = 0;
+				type = PathPointTypeStart;
+				g_array_append_val (points, pt);
+				g_byte_array_append (types, &type, 1);
+
+				type = PathPointTypeLine;
+				g_array_append_val (points, pt);
+				g_byte_array_append (types, &type, 1);
+
+				g_array_append_val (points, pt);
+				g_byte_array_append (types, &type, 1);
+
+				g_array_append_val (points, pt);
+				g_byte_array_append (types, &type, 1);
+				break;
+			}
+			/* beziers have 4 points: the previous one, the current and the next two */
+			i += 2;
+		} else {
+			/* no change required, just copy the point */
+			g_array_append_val (points, point);
+			g_byte_array_append (types, &type, 1);
+		}
+	}
+
+	/* free original path points and types */
+	if (path->points != NULL)
+		g_array_free (path->points, TRUE);
+	if (path->types != NULL)
+		g_byte_array_free (path->types, TRUE);
+
+	/* transfer new path informations */
+	path->points = points;
+	path->types = types;
+	path->count = points->len;
+
+	/* note: no error code is given for excessive recursion */
+	return Ok;
 }
 
 /* MonoTODO */
@@ -1106,8 +1359,11 @@ GdipTransformPath (GpPath* path, GpMatrix *matrix)
 
 	g_return_val_if_fail (path != NULL, InvalidParameter);
 
-	points = g_array_to_array (path->points);
 	count = path->count;
+	if (count == 0)
+		return Ok; /* GdipTransformMatrixPoints would fail */
+
+	points = g_array_to_array (path->points);
 
 	s = GdipTransformMatrixPoints (matrix, points, count);
 
@@ -1118,29 +1374,102 @@ GdipTransformPath (GpPath* path, GpMatrix *matrix)
         return s;
 }
 
-/* MonoTODO */
+/* MonoTODO - pen isn't considered (related to GdipWidenPath?) */
 GpStatus 
 GdipGetPathWorldBounds (GpPath *path, GpRectF *bounds, const GpMatrix *matrix, const GpPen *pen)
 {
-	static int called = 0;
+	GpStatus status = Ok;
+	GpPath *workpath = path;
 
-	if (!called) {
-		printf("NOT IMPLEMENTED: GdipGetPathWorldBounds (GpPath *path, GpRectF *bounds, const GpMatrix *matrix, const GpPen *pen)\n");
-		called = 1;
+	if (!path || !bounds)
+		return InvalidParameter;
+
+	if (matrix && !pen) {
+		status = GdipClonePath (path, &workpath);
+		if (status != Ok)
+			return status;
+		status = GdipTransformPath (workpath, (GpMatrix*)matrix);
+	} else if (pen) {
+		status = GdipClonePath (path, &workpath);
+		if (status != Ok)
+			return status;
+		/* FIXME: GdipWidenPath isn't implemented - and may not be related :-| */
+		status = GdipWidenPath (workpath, (GpPen*)pen, (GpMatrix*)matrix, 1.0f);
+	}
+
+	if (status != Ok) {
+		if (workpath != path)
+			GdipDeletePath (workpath);
+		return status;
+	} else {
+		int length = 0, i;
+		GpPointF *base, *points;
+
+		GdipGetPointCount (workpath, &length);
+		if (length < 1) {
+			/* special case #1 - Empty */
+			bounds->X = 0.0f;
+			bounds->Y = 0.0f;
+			bounds->Width = 0.0f;
+			bounds->Height = 0.0f;
+			return Ok;
+		}
+        
+		base = points = (GpPointF*) GdipAlloc (sizeof (GpPointF) * length);
+		GdipGetPathPoints (workpath, points, length);
+
+		bounds->X = points->X;		/* keep minimum X here */
+		bounds->Y = points->Y;		/* keep minimum Y here */
+		if (length == 1) {
+			/* special case #2 - Only one element */
+			bounds->Width = 0.0f;
+			bounds->Height = 0.0f;
+			GdipFree (base);
+			return Ok;
+		}
+
+		bounds->Width = points->X;	/* keep maximum X here */
+		bounds->Height = points->Y;	/* keep maximum Y here */
+		*points++;
+
+		for (i = 1; i < length; i++, *points++) {
+			if (points->X < bounds->X)
+				bounds->X = points->X;
+			if (points->Y < bounds->Y)
+				bounds->Y = points->Y;
+			if (points->X > bounds->Width)
+				bounds->Width = points->X;
+			if (points->Y > bounds->Height)
+				bounds->Height = points->Y;
+		}
+
+		GdipFree (base);
+
+		/* convert maximum values (width/height) as length */
+		bounds->Width -= bounds->X;
+		bounds->Height -= bounds->Y;
 	}
 	return Ok;
 }
 
-/* MonoTODO */
+/* MonoTODO - pen isn't considered (related to GdipWidenPath?) */
 GpStatus 
 GdipGetPathWorldBoundsI (GpPath *path, GpRect *bounds, const GpMatrix *matrix, const GpPen *pen)
 {
-	static int called = 0;
+	GpRectF rect;
+	GpStatus status;
 
-	if (!called) {
-		printf("NOT IMPLEMENTED: GdipGetPathWorldBoundsI (GpPath *path, GpRect *bounds, const GpMatrix *matrix, const GpPen *pen)\n");
-		called = 1;
-	}
+	if (!path || !bounds)
+		return InvalidParameter;
+
+	status = GdipGetPathWorldBounds (path, &rect, matrix, pen);
+	if (status != Ok)
+		return status;	
+
+	bounds->X = (int) rect.X;
+	bounds->Y = (int) rect.Y;
+	bounds->Width = (int) rect.Width;
+	bounds->Height = (int) rect.Height;
 	return Ok;
 }
 
