@@ -78,6 +78,7 @@ GdipDisposeImage (GpImage *image)
 {
 	int i = 0, j = 0, count = 0, dataCount = 0;
 	BitmapData *data;
+	bool dispose_bitmap;
 
 	if (!image)
 		return InvalidParameter;
@@ -87,33 +88,46 @@ GdipDisposeImage (GpImage *image)
 	image->surface = NULL;
 
 	count = image->frameDimensionCount;
+	dispose_bitmap = TRUE;
 
 	if (count > 0 && image->frameDimensionList != NULL) {
 		for (i = 0; i < count; i++) {
 			dataCount = image->frameDimensionList [i].count;
 			data = image->frameDimensionList [i].frames;
-				for (j = 0; j < dataCount; j++) {
-					if (((data [j].Reserved & GBD_OWN_SCAN0) != 0) && data [j].Scan0 != NULL) {
-						GdipFree (data [j].Scan0);
-						data [j].Scan0 = NULL;
-					}
-					if ((data [j].ByteCount) > 0 && (data [j].Bytes != NULL)) {
-						GdipFree (data [j].Bytes);
-						data [j].ByteCount = 0;
-						data [j].Bytes = NULL;
-					}
+			for (j = 0; j < dataCount; j++) {
+				if (((data [j].Reserved & GBD_OWN_SCAN0) != 0) && data [j].Scan0 != NULL) {
+					GdipFree (data [j].Scan0);
+					data [j].Scan0 = NULL;
 				}
+				if ((data [j].ByteCount) > 0 && (data [j].Bytes != NULL)) {
+					GdipFree (data [j].Bytes);
+					data [j].ByteCount = 0;
+					data [j].Bytes = NULL;
+				}
+			}
+			// FIXME: who's freeing 'data' ? (look around line 1551)
+			// someone, somewhere, is not playing fair with that (freeing will crash)
+			if (data != &((GpBitmap *)image)->data) {
+				GdipFree (data);
+			} else {
+				dispose_bitmap = FALSE;
+			}
 		}
 		GdipFree (image->frameDimensionList);
 	}
-	
-	/* Nothing more to be done here... We have already
-	 * cleaned the memory while looping in FrameDimension List
-	 * and hence we dont need to do anything in gdip_bitmap_dispose ()
-	 */
-	 GdipFree (image);
 
-	 return Ok;
+	if (image->palette) {
+		GdipFree (image->palette);
+		image->palette = NULL;
+	}
+	
+	if (dispose_bitmap) {
+		gdip_bitmap_dispose((GpBitmap *)image);
+	}
+
+	GdipFree (image);
+
+	return Ok;
 }
 
 /*
@@ -125,6 +139,7 @@ GdipGetImageGraphicsContext (GpImage *image, GpGraphics **graphics)
 {
 	GpGraphics *gfx;
 	cairo_surface_t *surface;
+	cairo_pattern_t *filter;
 	GpBitmap* bmp = (GpBitmap *) image;
 	
 	if (!image || !graphics)
@@ -137,23 +152,71 @@ GdipGetImageGraphicsContext (GpImage *image, GpGraphics **graphics)
 				bmp->data.Width, bmp->data.Height, bmp->data.Stride);
 
 	gfx = gdip_graphics_new (surface);
+	gfx->dpi_x = gfx->dpi_y = gdip_get_display_dpi ();
+	cairo_surface_destroy (surface);
 
 	gfx->image = bmp;
-	gfx->type = gtMemoryBitmap;	
-	cairo_pattern_set_filter (cairo_pattern_create_for_surface (bmp->image.surface), gdip_get_cairo_filter (gfx->interpolation));
+	gfx->type = gtMemoryBitmap;
+	filter = cairo_pattern_create_for_surface (bmp->image.surface);
+	cairo_pattern_set_filter (filter, gdip_get_cairo_filter (gfx->interpolation));
+	cairo_pattern_destroy (filter);
+	*graphics = gfx;
+	return Ok;
+}
+
+#ifdef CAIRO_HAS_PS_SURFACE
+
+#include <cairo-ps.h>
+
+GpStatus 
+GdipGetPostScriptGraphicsContext (char* filename, int width, int height, double dpix, double dpiy, GpGraphics **graphics)
+{
+	GpGraphics *gfx;
+	cairo_surface_t *surface;
+	
+	if (!graphics)
+		return InvalidParameter;
+
+	surface = cairo_ps_surface_create (filename, (double) width, (double) height);
+	cairo_ps_surface_set_dpi (surface, dpix, dpiy);
+
+	gfx = gdip_graphics_new (surface);
+	gfx->dpi_x = dpix;
+	gfx->dpi_y = dpiy;
+	cairo_surface_destroy (surface);
+
+	gfx->type = gtPostScript;
 	*graphics = gfx;
 	return Ok;
 }
 
+GpStatus 
+GdipGetPostScriptSavePage (GpGraphics* graphics) 
+{
+	if (!graphics)
+		return InvalidParameter;
+
+	cairo_show_page (graphics->ct);
+	return Ok;
+}
+
+#endif
+
 GpStatus 
 GdipDrawImageI (GpGraphics *graphics, GpImage *image, int x, int y)
 {
+	if (!image)
+		return InvalidParameter;
+
 	return GdipDrawImageRect (graphics, image, x, y, image->width, image->height);
 }
 
 GpStatus 
 GdipDrawImage (GpGraphics *graphics, GpImage *image, float x, float y)
 {
+	if (!image)
+		return InvalidParameter;
+
 	return GdipDrawImageRect (graphics, image, x, y, image->width, image->height);
 }
 
@@ -167,10 +230,17 @@ GpStatus
 GdipDrawImageRect (GpGraphics *graphics, GpImage *image, float x, float y, float width, float height)
 {
 	cairo_pattern_t *pattern;
+	cairo_pattern_t *org_pattern;
+
 	
 	g_return_val_if_fail (graphics != NULL, InvalidParameter);
 	g_return_val_if_fail (image != NULL, InvalidParameter);
 	g_return_val_if_fail (image->type == imageBitmap, InvalidParameter);
+
+	x = gdip_unitx_convgr (graphics, x);
+	y = gdip_unity_convgr (graphics, y);
+	width = gdip_unitx_convgr (graphics, width);
+	height = gdip_unity_convgr (graphics, height);
 	
 	cairo_new_path(graphics->ct);
 
@@ -193,6 +263,7 @@ GdipDrawImageRect (GpGraphics *graphics, GpImage *image, float x, float y, float
 	/* Create a surface for this bitmap if one doesn't exist */
 	gdip_bitmap_ensure_surface ((GpBitmap *) image);
 	pattern = cairo_pattern_create_for_surface (((GpBitmap*) image)->image.surface);
+
 	cairo_pattern_set_filter (pattern,
 				  gdip_get_cairo_filter (graphics->interpolation));
 
@@ -204,10 +275,14 @@ GdipDrawImageRect (GpGraphics *graphics, GpImage *image, float x, float y, float
 				(double) height / image->height);
 	}
 	
+	org_pattern = cairo_get_source(graphics->ct);
+	cairo_pattern_reference(org_pattern);
+
 	cairo_set_source_surface (graphics->ct, image->surface, 0, 0);
 	cairo_identity_matrix (graphics->ct);
 	cairo_paint (graphics->ct);
-	
+	cairo_set_source(graphics->ct, org_pattern);	
+
 	cairo_pattern_destroy (pattern);
 	
 	return Ok;
@@ -219,6 +294,7 @@ GdipDrawImagePoints (GpGraphics *graphics, GpImage *image, GDIPCONST GpPointF *d
 {
 	float width, height;
 	cairo_pattern_t *pattern;
+	cairo_pattern_t *org_pattern;
 	
 	g_return_val_if_fail (graphics != NULL, InvalidParameter);
 	g_return_val_if_fail (image != NULL, InvalidParameter);
@@ -261,9 +337,14 @@ GdipDrawImagePoints (GpGraphics *graphics, GpImage *image, GDIPCONST GpPointF *d
 		(double) width / image->width,
 		(double) height / image->height);
 	
+	org_pattern = cairo_get_source(graphics->ct);
+	cairo_pattern_reference(org_pattern);
+
 	cairo_set_source_surface (graphics->ct, image->surface, 0, 0);
 	cairo_identity_matrix (graphics->ct);
 	cairo_paint (graphics->ct);	
+	cairo_set_source(graphics->ct, org_pattern);
+
 	cairo_pattern_destroy (pattern);
 	
 	return Ok;
@@ -292,7 +373,11 @@ GdipDrawImagePointRect (GpGraphics *graphics, GpImage *image,
                         float srcx, float srcy, float srcwidth, float srcheight,
                         GpUnit srcUnit)
 {
-	return NotImplemented; /* GdipDrawImagePointRect */
+	if (!image)
+		return InvalidParameter;
+
+	return GdipDrawImageRectRect (graphics, image, x, y, srcwidth, srcheight, srcx, srcy, 
+		srcwidth, srcheight, srcUnit, NULL, NULL, NULL);
 }
 
 GpStatus
@@ -313,6 +398,7 @@ GdipDrawImageRectRect (GpGraphics *graphics, GpImage *image,
                        DrawImageAbort callback, void *callbackData)
 {
 	cairo_pattern_t *pattern;
+	cairo_pattern_t *orig;
 	cairo_matrix_t mat;
 	void* dest, *org;
 	bool allocated = FALSE;
@@ -344,14 +430,14 @@ GdipDrawImageRectRect (GpGraphics *graphics, GpImage *image,
 	}
 
 	if (srcUnit != UnitPixel && srcUnit != UnitWorld) {
-		gdip_unitConversion(srcUnit, UnitPixel, dstx, &dstx);
-		gdip_unitConversion(srcUnit, UnitPixel, dsty, &dsty);
-		gdip_unitConversion(srcUnit, UnitPixel, dstwidth, &dstwidth);
-		gdip_unitConversion(srcUnit, UnitPixel, dstheight, &dstheight);
-		gdip_unitConversion(srcUnit, UnitPixel, srcx, &srcx);
-		gdip_unitConversion(srcUnit, UnitPixel, srcy, &srcy);
-		gdip_unitConversion(srcUnit, UnitPixel, srcwidth, &dstwidth);
-		gdip_unitConversion(srcUnit, UnitPixel, srcheight, &srcheight);
+		gdip_unit_conversion (srcUnit, UnitCairoPoint, graphics->type, graphics->dpi_x, dstx, &dstx);
+		gdip_unit_conversion (srcUnit, UnitCairoPoint, graphics->type, graphics->dpi_y, dsty, &dsty);
+		gdip_unit_conversion (srcUnit, UnitCairoPoint, graphics->type, graphics->dpi_x, dstwidth, &dstwidth);
+		gdip_unit_conversion (srcUnit, UnitCairoPoint, graphics->type, graphics->dpi_y, dstheight, &dstheight);
+		gdip_unit_conversion (srcUnit, UnitCairoPoint, graphics->type, graphics->dpi_x, srcx, &srcx);
+		gdip_unit_conversion (srcUnit, UnitCairoPoint, graphics->type, graphics->dpi_y, srcy, &srcy);
+		gdip_unit_conversion (srcUnit, UnitCairoPoint, graphics->type, graphics->dpi_x, srcwidth, &dstwidth);
+		gdip_unit_conversion (srcUnit, UnitCairoPoint, graphics->type, graphics->dpi_y, srcheight, &srcheight);
 	}
 
 	org = dest = bitmap->data.Scan0; 
@@ -364,7 +450,7 @@ GdipDrawImageRectRect (GpGraphics *graphics, GpImage *image,
 		bitmap->data.Scan0 = dest;
 	
 	/* Re-attach the image if the image has attribues */
-	if ((bitmap->image.surface && imageAttributes != NULL) || bitmap->image.surface == NULL)  {
+	if ((bitmap->image.surface != NULL) && (imageAttributes != NULL))  {
 		cairo_surface_destroy (bitmap->image.surface);
 		bitmap->image.surface = NULL;
 	}
@@ -381,6 +467,8 @@ GdipDrawImageRectRect (GpGraphics *graphics, GpImage *image,
 			flipXOn = flipYOn = TRUE;
 		
 		if (flipXOn) {			
+			// We're ok just cloning the bitmap, we don't need the image data
+			// and we destroy it before we leave this function
 			gdip_bitmap_clone (bitmap, &imgflipX);
 			gdip_flip_x ((GpImage *) imgflipX);	
 			gdip_bitmap_ensure_surface (imgflipX);			
@@ -401,7 +489,6 @@ GdipDrawImageRectRect (GpGraphics *graphics, GpImage *image,
 		
 		/* Create a surface for this bitmap if one doesn't exist */
 		gdip_bitmap_ensure_surface ((GpBitmap*) image);
-		cairo_pattern_set_filter (cairo_pattern_create_for_surface(((GpBitmap*) image)->image.surface), gdip_get_cairo_filter (graphics->interpolation));
 
 		for (posy = 0; posy < dstheight; posy += img_height) {
 		
@@ -424,11 +511,16 @@ GdipDrawImageRectRect (GpGraphics *graphics, GpImage *image,
 
 				pattern = cairo_pattern_create_for_surface(cur_image->image.surface);
 				cairo_pattern_set_matrix (pattern, &mat);
+
+				orig = cairo_get_source(graphics->ct);
+				cairo_pattern_reference(orig);
+
 				cairo_set_source(graphics->ct, pattern);
-				
 				cairo_rectangle (graphics->ct, dstx + posx, dsty + posy, img_width, img_height);
 				cairo_fill (graphics->ct);
-				
+
+				cairo_set_source(graphics->ct, orig);
+
 				cairo_matrix_init_identity (&mat);
 				cairo_pattern_set_matrix (pattern, &mat);
 
@@ -453,24 +545,33 @@ GdipDrawImageRectRect (GpGraphics *graphics, GpImage *image,
 			GdipDisposeImage ((GpImage *) imgflipXY);
 		
 	}
-	else  {		
+	else {
+		cairo_pattern_t *filter;
+
 		/* Create a surface for this bitmap if one doesn't exist */
 		gdip_bitmap_ensure_surface ((GpBitmap*) image);
-		cairo_pattern_set_filter (cairo_pattern_create_for_surface (((GpBitmap*) image)->image.surface) , gdip_get_cairo_filter (graphics->interpolation));
+		filter = cairo_pattern_create_for_surface (((GpBitmap*) image)->image.surface);
+		cairo_pattern_set_filter (filter, gdip_get_cairo_filter (graphics->interpolation));
 		
 		cairo_matrix_scale (&mat, srcwidth / dstwidth, srcheight / dstheight);
 		cairo_matrix_translate (&mat, srcx - dstx, srcy - dsty);
 
 		pattern = cairo_pattern_create_for_surface(image->surface);
 		cairo_pattern_set_matrix (pattern, &mat);
+
+		orig = cairo_get_source(graphics->ct);
+		cairo_pattern_reference(orig);
+
 		cairo_set_source(graphics->ct, pattern);
-		
 		cairo_rectangle (graphics->ct, dstx, dsty, dstwidth, dstheight);
 		cairo_fill (graphics->ct);
 		
+		cairo_set_source(graphics->ct, orig);
+
 		cairo_matrix_init_identity (&mat);
 		cairo_pattern_set_matrix (pattern, &mat);
 		cairo_pattern_destroy (pattern);
+		cairo_pattern_destroy (filter);
 	}
 
 	/* The default surface is no longer valid*/
@@ -636,8 +737,9 @@ gdip_get_imageformat_from_codec_clsid (CLSID *encoderCLSID)
 
     	for (cnt = 0, encoder = encoders; cnt < numEncoders; encoder++) {
 	       	if (memcmp (&encoder->Clsid, encoderCLSID, sizeof (GUID)) == 0) {
+			ImageFormat ifmt = gdip_image_format_for_format_guid (&encoder->FormatID);
             		GdipFree (encoders);
-            		return gdip_image_format_for_format_guid (&encoder->FormatID);
+			return ifmt;
         	}
     	}
 
@@ -836,7 +938,7 @@ GdipGetImagePixelFormat (GpImage *image, PixelFormat *format)
 GpStatus 
 GdipImageGetFrameDimensionsCount (GpImage *image, UINT *count)
 {
-	if (!image)
+	if (!image || !count)
                 return InvalidParameter;
 
 	*count = image->frameDimensionCount;
@@ -870,7 +972,7 @@ GdipImageGetFrameCount(GpImage *image, GDIPCONST GUID *dimensionGUID, UINT* coun
 {
 	int i;
 
-	if (!image || !dimensionGUID)
+	if (!image || !dimensionGUID || !count)
 		return InvalidParameter;
 
 	for (i=0; i<image->frameDimensionCount; i++){
@@ -1400,7 +1502,10 @@ GdipGetImagePaletteSize (GpImage *image, int* size)
 GpStatus 
 GdipGetPropertyCount (GpImage *image, UINT *propertyNumber)
 {
-	return NotImplemented;
+	/*
+	 * We currently only return 1 fake property 0xdeadbeee
+	 */
+	return 1;
 }
 
 GpStatus 
@@ -1412,31 +1517,42 @@ GdipGetPropertyIdList (GpImage *image, UINT propertyNumber, PROPID *list)
 GpStatus 
 GdipGetPropertyItemSize (GpImage *image, PROPID propID, UINT *size)
 {
-	return NotImplemented;
+	*size = 16;
+	return Ok;
 }
 
 GpStatus 
 GdipGetPropertyItem (GpImage *image, PROPID propID, UINT size, PropertyItem buffer)
 {
-	return NotImplemented;
+	if (propID != 0xdeadbeee)
+		return InvalidParameter;
+
+	return Ok;
 }
 
 GpStatus 
 GdipGetPropertySize (GpImage *image, UINT *bufferSize, UINT *propertyNumbers)
 {
-	return NotImplemented;
+	if (!image || !bufferSize || !propertyNumbers)
+		return InvalidParameter;
+
+	*bufferSize = 32;
+	*propertyNumbers = 1;
+	return Ok;
 }
 
 GpStatus 
 GdipRemovePropertyItem (GpImage *image, PROPID propID)
 {
-	return NotImplemented;
+	/* Not implemented */
+	return Ok;
 }
 
 GpStatus 
 GdipSetPropertyItem(GpImage *image, GDIPCONST PropertyItem *item)
 {
-	return NotImplemented; /* GdipSetPropertyItem */
+	/* Not implemented */
+	return Ok;
 }
 
 void
@@ -1476,6 +1592,15 @@ gdip_image_clone (GpImage* image, GpImage* clonedImage)
 			}
 		}
 	}
+	// if present, clone the palette too
+	if (image->palette) {
+		int psize;
+		if (GdipGetImagePaletteSize (image, &psize) == Ok) {
+			clonedImage->palette = GdipAlloc (psize);
+			memcpy (clonedImage->palette, image->palette, psize);
+		} else
+			clonedImage->palette = NULL;
+	}
 }
 
 GpStatus
@@ -1854,7 +1979,15 @@ GdipGetImageEncoders (UINT numEncoders, UINT size, ImageCodecInfo *encoders)
 GpStatus
 GdipGetAllPropertyItems(GpImage *image, UINT totalBufferSize, UINT numProperties, PropertyItem *allItems)
 {
-	return(NotImplemented);
+	if (!allItems || (totalBufferSize < 16))
+		return InvalidParameter;
+
+	allItems [0].id = 0xdeadbeee;
+	allItems [0].length = 4;
+	/* A byte */
+	allItems [0].type = 1;
+	allItems [0].value = &(allItems [0].id);
+	return Ok;
 }
 
 GpStatus
