@@ -111,6 +111,7 @@ gdip_linear_gradient_clone_brush (GpBrush *brush, GpBrush **clonedBrush)
 	if (newbrush->presetColors == NULL) 
 		goto NO_PRESET;
 
+	newbrush->presetColors->count = linear->presetColors->count;
 	if (linear->presetColors->count > 0) {
 		newbrush->presetColors->colors = (ARGB *) GdipAlloc (linear->presetColors->count * sizeof (ARGB));
 		if (newbrush->presetColors->colors == NULL) 
@@ -173,29 +174,38 @@ gdip_linear_gradient_destroy (GpBrush *brush)
 
 	linear = (GpLineGradient *) brush;
 
-	if (linear->rectangle)
+	if (linear->rectangle) {
 		GdipFree (linear->rectangle);
-	linear->rectangle = NULL;
+		linear->rectangle = NULL;
+	}
 
-	if (linear->blend)
+	if (linear->blend) {
 		if (linear->blend->count > 0) {
 			GdipFree (linear->blend->factors);
 			GdipFree (linear->blend->positions);
 		}
-	GdipFree (linear->blend);
-	linear->blend = NULL;
+		GdipFree (linear->blend);
+		linear->blend = NULL;
+	}
 
-	if (linear->presetColors)
+	if (linear->presetColors) {
 		if (linear->presetColors->count > 0) {
 			GdipFree (linear->presetColors->colors);
 			GdipFree (linear->presetColors->positions);
 		}
-	GdipFree (linear->presetColors);
-	linear->presetColors = NULL;
+		GdipFree (linear->presetColors);
+		linear->presetColors = NULL;
+	}
 
-	if (linear->pattern)
+	if (linear->pattern) {
 		cairo_pattern_destroy (linear->pattern);
-	linear->pattern = NULL;
+		linear->pattern = NULL;
+	}
+
+	if (linear->matrix) {
+		GdipDeleteMatrix (linear->matrix);
+		linear->matrix = NULL;
+	}
 
 	GdipFree (linear);
 
@@ -289,10 +299,8 @@ create_tile_linear (cairo_t *ct, GpLineGradient *linear)
 	g_return_val_if_fail (rect != NULL, InvalidParameter);
 
 	/* gradient surface to be created. We are using CAIRO_CONTENT_COLOR_ALPHA, */
-	/* because internally this is the format we use everywhere. */
 	gradient = cairo_surface_create_similar (cairo_get_target (ct), CAIRO_CONTENT_COLOR_ALPHA,
 						 (int) (2 * rect->Width), (int) (2 * rect->Height));
-
 	if (gradient == NULL) {
 		return OutOfMemory;
 	}
@@ -342,7 +350,7 @@ create_tile_linear (cairo_t *ct, GpLineGradient *linear)
 
 	linear->pattern = cairo_pattern_create_for_surface (gradient);
 	cairo_surface_destroy (gradient);
-
+	GdipDeleteMatrix (tempMatrix);
 
 	return Ok;
 }
@@ -358,7 +366,6 @@ create_tile_flipX_linear (cairo_t *ct, GpLineGradient *linear)
 	g_return_val_if_fail (rect != NULL, InvalidParameter);
 
 	/* gradient surface to be created. We are using CAIRO_CONTENT_COLOR_ALPHA, */
-	/* because internally this is the format we use everywhere. */
 	gradient = cairo_surface_create_similar (cairo_get_target (ct), CAIRO_CONTENT_COLOR_ALPHA,
 						 (int) (2 * rect->Width), (int) (2 * rect->Height));
 
@@ -379,6 +386,7 @@ create_tile_flipX_linear (cairo_t *ct, GpLineGradient *linear)
 	GdipCreateMatrix (&tempMatrix);
 	cairo_save (ct);
 	{
+		cairo_t *temp = ct;
 		ct = cairo_create (gradient);
 
 		if (linear->blend->count > 1)
@@ -405,6 +413,8 @@ create_tile_flipX_linear (cairo_t *ct, GpLineGradient *linear)
 		cairo_fill (ct);
 
 		cairo_pattern_destroy (pat);
+		cairo_destroy (ct);
+		ct = temp;
 	}
 	cairo_restore (ct);
 
@@ -431,8 +441,6 @@ create_tile_flipX_linear (cairo_t *ct, GpLineGradient *linear)
 	cairo_matrix_multiply (tempMatrix, tempMatrix, currMatrix);
 	cairo_set_matrix (ct, tempMatrix);
 	GdipDeleteMatrix (currMatrix);
-	
-	
 
 	return Ok;
 }
@@ -448,7 +456,6 @@ create_tile_flipY_linear (cairo_t *ct, GpLineGradient *linear)
 	g_return_val_if_fail (rect != NULL, InvalidParameter);
 
 	/* gradient surface to be created. We are using CAIRO_CONTENT_COLOR_ALPHA, */
-	/* because internally this is the format we use everywhere. */
 	gradient = cairo_surface_create_similar (cairo_get_target (ct), CAIRO_CONTENT_COLOR_ALPHA,
 						 (int) (2 * rect->Width), (int) (2 * rect->Height));
 
@@ -469,6 +476,7 @@ create_tile_flipY_linear (cairo_t *ct, GpLineGradient *linear)
 	GdipCreateMatrix (&tempMatrix);	
 	cairo_save (ct);
 	{
+		cairo_t *temp = ct;
 		ct = cairo_create (gradient);
 
 		if (linear->blend->count > 1)
@@ -495,6 +503,8 @@ create_tile_flipY_linear (cairo_t *ct, GpLineGradient *linear)
 		cairo_fill (ct);
 
 		cairo_pattern_destroy (pat);
+		cairo_destroy (ct);
+		ct = temp;
 	}
 	cairo_restore (ct);
 
@@ -538,7 +548,6 @@ create_tile_flipXY_linear (cairo_t *ct, GpLineGradient *linear)
 	g_return_val_if_fail (rect != NULL, InvalidParameter);
 
 	/* gradient surface to be created. We are using CAIRO_CONTENT_COLOR_ALPHA, */
-	/* because internally this is the format we use everywhere. */
 	gradient = cairo_surface_create_similar (cairo_get_target (ct), CAIRO_CONTENT_COLOR_ALPHA,
 						 (int) (2 * rect->Width), (int) (2 * rect->Height));
 	
@@ -559,6 +568,7 @@ create_tile_flipXY_linear (cairo_t *ct, GpLineGradient *linear)
 	GdipCreateMatrix (&tempMatrix);	
 	cairo_save (ct);
 	{
+		cairo_t *temp = ct;
 		ct = cairo_create (gradient);
 
 		if (linear->blend->count > 1)
@@ -615,6 +625,8 @@ create_tile_flipXY_linear (cairo_t *ct, GpLineGradient *linear)
 		cairo_fill (ct);
 
 		cairo_pattern_destroy (pat);
+		cairo_destroy (ct);
+		ct = temp;
 	}
 	cairo_restore (ct);
 
@@ -704,6 +716,7 @@ gdip_linear_gradient_setup (GpGraphics *graphics, GpBrush *brush)
 			cairo_set_source (ct, linear->pattern);
 
 			status = gdip_get_status (cairo_status (ct));
+			GdipDeleteMatrix (product);
 		}
 		else
 			status = GenericError;
@@ -740,33 +753,17 @@ gdip_set_rect (GpRectF *rect, float x1, float y1, float x2, float y2)
 GpStatus
 GdipCreateLineBrushI (GDIPCONST GpPoint *point1, GDIPCONST GpPoint *point2, ARGB color1, ARGB color2, GpWrapMode wrapMode, GpLineGradient **lineGradient)
 {
-	GpPointF *pt1, *pt2;
-	GpStatus status;
+	GpPointF pt1, pt2;
 
 	g_return_val_if_fail (point1 != NULL, InvalidParameter);
 	g_return_val_if_fail (point2 != NULL, InvalidParameter);
 
-	pt1 = (GpPointF *) GdipAlloc (sizeof (GpPointF));
-	g_return_val_if_fail (pt1 != NULL, OutOfMemory);
+	pt1.X = point1->X; 
+	pt1.Y = point1->Y;
+	pt2.X = point2->X; 
+	pt2.Y = point2->Y;
 
-	pt2 = (GpPointF *) GdipAlloc (sizeof (GpPointF));
-
-	if (pt2 == NULL) {
-		GdipFree (pt1);
-		return OutOfMemory;
-	}
-
-	pt1->X = point1->X; 
-	pt1->Y = point1->Y;
-	pt2->X = point2->X; 
-	pt2->Y = point2->Y;
-
-	status = GdipCreateLineBrush (pt1, pt2, color1, color2, wrapMode, lineGradient);
-
-	GdipFree (pt1);
-	GdipFree (pt2);
-
-	return status;
+	return GdipCreateLineBrush (&pt1, &pt2, color1, color2, wrapMode, lineGradient);
 }
 
 GpStatus
@@ -793,8 +790,10 @@ GdipCreateLineBrush (GDIPCONST GpPointF *point1, GDIPCONST GpPointF *point2, ARG
 	linear->wrapMode = wrapMode;
 	linear->lineColors [0] = color1;
 	linear->lineColors [1] = color2;
-	linear->points [0] = *point1;
-	linear->points [1] = *point2;
+	linear->points [0].X = point1->X;
+	linear->points [0].Y = point1->Y;
+	linear->points [1].X = point2->X;
+	linear->points [1].Y = point2->Y;
 	linear->rectangle = rectf;
 
 	*lineGradient = linear;
@@ -859,19 +858,16 @@ GdipCreateLineBrushFromRect (GDIPCONST GpRectF *rect, ARGB color1, ARGB color2,
 GpStatus
 GdipCreateLineBrushFromRectWithAngleI (GDIPCONST GpRect *rect, ARGB color1, ARGB color2, float angle, BOOL isAngleScalable, GpWrapMode wrapMode, GpLineGradient **lineGradient)
 {
-	GpRectF *rectf;
+	GpRectF rectf;
 
 	g_return_val_if_fail (rect != NULL, InvalidParameter);
 
-	rectf = (GpRectF *) GdipAlloc (sizeof (GpRectF));
-	g_return_val_if_fail (rectf != NULL, OutOfMemory);
+	rectf.X = rect->X;
+	rectf.Y = rect->Y;
+	rectf.Width = rect->Width;
+	rectf.Height = rect->Height;
 
-	rectf->X = rect->X;
-	rectf->Y = rect->Y;
-	rectf->Width = rect->Width;
-	rectf->Height = rect->Height;
-
-	return GdipCreateLineBrushFromRectWithAngle (rectf, color1, color2, angle, 
+	return GdipCreateLineBrushFromRectWithAngle (&rectf, color1, color2, angle, 
 						     isAngleScalable, wrapMode, lineGradient);
 }
 
@@ -1162,7 +1158,7 @@ GdipGetLineTransform (GpLineGradient *brush, GpMatrix *matrix)
 	if (brush->presetColors->count >= 2)
 		return WrongState;
 
-	gdip_cairo_matrix_copy(matrix, &brush->matrix);
+	gdip_cairo_matrix_copy (matrix, brush->matrix);
 
 	return Ok;
 }
@@ -1173,7 +1169,7 @@ GdipSetLineTransform (GpLineGradient *brush, GDIPCONST GpMatrix *matrix)
 	g_return_val_if_fail (brush != NULL, InvalidParameter);
 	g_return_val_if_fail (matrix != NULL, InvalidParameter);
 
-	gdip_cairo_matrix_copy(&brush->matrix, matrix);
+	gdip_cairo_matrix_copy (brush->matrix, matrix);
 	brush->base.changed = TRUE;
 	return Ok;
 }
@@ -1449,14 +1445,17 @@ GdipSetLineSigmaBlend (GpLineGradient *brush, float focus, float scale)
 GpStatus
 GdipMultiplyLineTransform (GpLineGradient *brush, GpMatrix *matrix, GpMatrixOrder order)
 {
-	GpStatus status;
-
 	g_return_val_if_fail (brush != NULL, InvalidParameter);
 	g_return_val_if_fail (matrix != NULL, InvalidParameter);
 
-	if ((status = GdipMultiplyMatrix (brush->matrix, matrix, order)) == Ok)
-		brush->base.changed = TRUE;
-	return status;
+	/* note: error handling is different from GdipMultiplyMatrix */
+	if (order == MatrixOrderAppend)
+		cairo_matrix_multiply (brush->matrix, brush->matrix, matrix);
+	else
+		cairo_matrix_multiply (brush->matrix, matrix, brush->matrix);        
+
+	brush->base.changed = TRUE;
+	return Ok;
 }
 
 GpStatus
