@@ -48,8 +48,7 @@ gdip_pen_init (GpPen *pen)
 	pen->compound_array = NULL;
 	pen->unit = UnitWorld;
 	pen->changed = TRUE;
-	pen->matrix = NULL;
-	pen->matrix = g_new (GpMatrix, 1);
+	GdipCreateMatrix (&pen->matrix);
 	cairo_matrix_init_identity (pen->matrix);
 }
 
@@ -129,13 +128,13 @@ gdip_pen_setup (GpGraphics *graphics, GpPen *pen)
 	g_return_val_if_fail (graphics != NULL, InvalidParameter);
 	g_return_val_if_fail (pen != NULL, InvalidParameter);
 
-	GdipCreateMatrix (&product);
-	cairo_matrix_init_identity (product);
-	
 	status = gdip_brush_setup (graphics, pen->brush);
 	if (status != Ok)
 		return status;
 
+	GdipCreateMatrix (&product);
+	cairo_matrix_init_identity (product);
+
 	/* Here we use product of pen->matrix and graphics->copy_of_ctm.
 	 * This gives us absolute results with respect to graphics. We
 	 * do following irrespective of the pen->changed state since graphics
@@ -146,6 +145,8 @@ gdip_pen_setup (GpGraphics *graphics, GpPen *pen)
 	cairo_matrix_multiply (product, pen->matrix, graphics->copy_of_ctm);
 	cairo_set_matrix (graphics->ct, product);
 
+	GdipDeleteMatrix (product);
+
 	/* Don't need to setup, if pen is the same as the cached pen and
 	 * it is not changed. Just comparing pointers may not be sufficient
 	 * to say that the pens are same. It is possible to have different
@@ -335,7 +336,9 @@ GdipClonePen (GpPen *pen, GpPen **clonepen)
 	result->compound_count = pen->compound_count;
 	result->compound_array = compound_array;
 	result->unit = pen->unit;
-        gdip_cairo_matrix_copy(result->matrix, matrix);
+	if (result->matrix)
+		GdipDeleteMatrix (result->matrix);
+	result->matrix = matrix;
 	result->changed = pen->changed;
 
         *clonepen = result;
@@ -354,12 +357,21 @@ GdipDeletePen (GpPen *pen)
 		pen->dash_array = NULL;
 	}
 
-	if (pen->own_brush && pen->brush)
+	if (pen->own_brush && pen->brush) {
 		GdipFree (pen->brush);
+		pen->brush = NULL;
+	}
 
-        if (pen->compound_count != 0)
+        if (pen->compound_count != 0) {
                 GdipFree (pen->compound_array);
-	pen->compound_array = NULL;
+		pen->compound_array = NULL;
+		pen->compound_count = 0;
+	}
+
+	if (pen->matrix) {
+		GdipDeleteMatrix (pen->matrix);
+		pen->matrix = NULL;
+	}
 
         GdipFree (pen);
 
