@@ -1,5 +1,6 @@
 /*
  * Copyright (c) 2004-2005 Ximian
+ * Copyright (C) 2006 Novell, Inc (http://www.novell.com)
  *
  * Permission is hereby granted, free of charge, to any person obtaining a copy of this software
  * and associated documentation files (the "Software"), to deal in the Software without restriction,
@@ -16,13 +17,14 @@
  * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
  * OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
  *
- * Author:
+ * Authors:
  *          Jordi Mas i Hernandez <jordi@ximian.com>, 2004-2005
+ *          Sebastien Pouliot  <sebastien@ximian.com>
  *
  */
 
-#include "gdip.h"
-
+#include "region.h"
+#include "matrix.h"
 
 /*
 	Helper functions
@@ -81,16 +83,62 @@ gdip_is_Point_in_RectFs_Visible (float x, float y, GpRectF* r, int cnt)
         return FALSE;
 }
 
-bool
-gdip_is_InfiniteRegion (GpRegion *region)
-{
-	if (region->cnt != 1)
-              return FALSE;
+void gdip_get_bounds (GpRectF *allrects, int allcnt, GpRectF *bound);
 
-	if (region->rects->X == -4194304 && region->rects->Y == -4194304 &&
-		region->rects->Width == 8388608 &&  region->rects->Height == 8388608)
+/* This internal version doesn't require a Graphic object to work */
+static BOOL
+gdip_is_region_empty (GpRegion *region)
+{
+	GpRectF rect;
+
+	if (!region)
+		return FALSE;
+
+	if (region->type == RegionTypePath) {
+		/* check for an existing, but empty, path list */
+		if (!region->tree)
+			return TRUE;
+		if (region->tree->path)
+			return (region->tree->path->count == 0);
+		return FALSE;
+	}
+
+	if (!region->rects || (region->cnt == 0))
 		return TRUE;
 
+	gdip_get_bounds (region->rects, region->cnt, &rect);
+	return ((rect.Width == 0) || (rect.Height == 0));
+}
+
+static BOOL
+gdip_is_rect_infinite (GpRectF *rect)
+{
+	return ((rect->X == REGION_INFINITE_POSITION) && 
+		(rect->Y == REGION_INFINITE_POSITION) &&
+		(rect->Width == REGION_INFINITE_LENGTH) && 
+		(rect->Height == REGION_INFINITE_LENGTH));
+}
+
+BOOL
+gdip_is_InfiniteRegion (GpRegion *region)
+{
+	switch (region->type) {
+	case RegionTypeRectF:
+		if (region->cnt != 1)
+	              return FALSE;
+		return gdip_is_rect_infinite (region->rects);
+	case RegionTypePath:
+		/* FIXME: incomplete and not 100% accurate (curves) - but cover the most common case */
+		if (region->tree && region->tree->path && (region->tree->path->count == 4)) {
+			GpRectF bounds;
+			if (GdipGetPathWorldBounds (region->tree->path, &bounds, NULL, NULL) != Ok)
+				return gdip_is_rect_infinite (&bounds);
+		}
+		break;
+	default:
+		g_warning ("unknown type %d", region->type);
+		break;
+	}
 	return FALSE;
 }
 
@@ -191,6 +239,94 @@ gdip_getlowestrect (GpRectF *rects, int cnt, GpRectF* src, GpRectF* rslt)
 	return TRUE;
 }
 
+void 
+gdip_clear_region (GpRegion *region)
+{
+	region->type = RegionTypeEmpty;
+
+        if (region->rects) {
+                GdipFree (region->rects);
+		region->rects = NULL;
+	}
+
+	if (region->tree) {
+		gdip_region_clear_tree (region->tree);
+                GdipFree (region->tree);
+		region->tree = NULL;
+	}
+
+	if (region->bitmap) {
+		gdip_region_bitmap_free (region->bitmap);
+		region->bitmap = NULL;
+	}
+
+	region->cnt = 0;
+}
+
+void
+gdip_copy_region (GpRegion *source, GpRegion *dest)
+{
+	dest->type = source->type;
+
+	if (source->rects) {
+		dest->cnt = source->cnt;
+	        dest->rects = (GpRectF *) GdipAlloc (sizeof (GpRectF) * source->cnt);
+        	memcpy (dest->rects, source->rects, sizeof (GpRectF) * source->cnt);
+	} else {
+		dest->rects = NULL;
+	}
+
+	if (source->tree) {
+		dest->tree = (GpPathTree *) GdipAlloc (sizeof (GpPathTree));
+		gdip_region_copy_tree (source->tree, dest->tree);
+	} else {
+		dest->tree = NULL;
+	}
+
+	if (source->bitmap) {
+		dest->bitmap = gdip_region_bitmap_clone (source->bitmap);
+	} else {
+		dest->bitmap = NULL;
+	}
+}
+
+/* convert a rectangle-based region to a path based region */
+static void 
+gdip_region_convert_to_path (GpRegion *region)
+{
+	int i;
+	GpRectF *rect;
+
+	/* no convertion is required for empty or complex regions */
+	if (!region || (region->cnt == 0) || (region->type != RegionTypeRectF))
+		return;
+
+	region->type = RegionTypePath;
+	region->tree = (GpPathTree *) GdipAlloc (sizeof (GpPathTree));
+
+	GdipCreatePath (FillModeAlternate, &region->tree->path);
+	/* all rectangles are converted into a single path */
+	for (i = 0, rect = region->rects; i < region->cnt; i++, rect++) {
+		GdipAddPathRectangle (region->tree->path, rect->X, rect->Y, rect->Width, rect->Height);
+	}
+
+	if (region->rects) {
+		GdipFree (region->rects);
+		region->rects = NULL;
+	}
+}
+
+/*
+ * Create a region (path-tree) from a path.
+ */
+static GpStatus
+gdip_region_create_from_path (GpRegion *region, GpPath *path)
+{
+	region->type = RegionTypePath;
+	region->tree = (GpPathTree *) GdipAlloc (sizeof (GpPathTree));
+	return GdipClonePath (path, &region->tree->path);
+}
+
 GpStatus
 gdip_createRegion (GpRegion **region, RegionType type, void *src)
 {
@@ -198,25 +334,31 @@ gdip_createRegion (GpRegion **region, RegionType type, void *src)
         GpRectF rect;
 
         result = (GpRegion *) GdipAlloc (sizeof (GpRegion));
+	result->type = type;
+	result->cnt = 0;
         result->rects = NULL;
-        result->cnt = 0;
+        result->tree = NULL;
+        result->bitmap = NULL;
 
         switch (type) {
         case RegionTypeRect:
                 gdip_from_Rect_To_RectF ((GpRect *)src, &rect);
                 gdip_add_rect_to_array (&result->rects, &result->cnt,  &rect);
+		result->type = RegionTypeRectF;
                 break;
         case RegionTypeRectF:
                 gdip_add_rect_to_array (&result->rects, &result->cnt,  (GpRectF *)src);
                 break;
         case RegionTypeEmpty:
 		GdipSetInfinite (result);
+		/* note: GdipSetInfinite converts type to RegionTypeRectF */
                 break;
-
+	case RegionTypePath:
+		gdip_region_create_from_path (result, (GpPath*)src);
+		break;
         default:
-                GdipFree (result);
-                return NotImplemented;
-
+		g_warning ("unknown type %d", result->type);
+		return NotImplemented;
         }
 
 	*region = result;
@@ -256,6 +398,51 @@ GdipCreateRegionRectI (GDIPCONST GpRect *rect, GpRegion **region)
         return gdip_createRegion (region, RegionTypeRect, (void*) rect);
 }
 
+GpStatus
+GdipCreateRegionRgnData (GDIPCONST BYTE *regionData, int size, GpRegion **region)
+{
+	GpRegion *result;
+
+	if (!region || !regionData || (size < 8))
+		return InvalidParameter;
+
+	result = (GpRegion *) GdipAlloc (sizeof (GpRegion));
+	result->type = *(guint32*) regionData;
+	result->cnt = 0;
+        result->rects = NULL;
+	result->tree = NULL;
+        result->bitmap = NULL;
+
+	switch (result->type) {
+	case RegionTypeRectF: {
+		guint32 count = (guint32) *(regionData + 4);
+		GpRectF *rect;
+		int i;
+
+		if (count != (size - 8) / sizeof (GpRectF))
+			return InvalidParameter;
+
+		for (i = 0, rect = (GpRectF*)(regionData + 8); i < count; i++, rect++)
+	                gdip_add_rect_to_array (&result->rects, &result->cnt, rect);
+		}
+		break;
+	case RegionTypePath: {
+		if (size < 16)
+			return InvalidParameter;
+
+		result->tree = (GpPathTree*) GdipAlloc (sizeof (GpPathTree));
+		if (!gdip_region_deserialize_tree ((BYTE*)(regionData + 4), (size - 4), result->tree))
+			return InvalidParameter;
+		}
+		break;
+	default:
+		g_warning ("unknown type %d", result->type);
+		return NotImplemented;
+	}
+
+	*region = result;
+	return Ok;
+}
 
 GpStatus
 GdipCloneRegion (GpRegion *region, GpRegion **cloneRegion)
@@ -266,9 +453,7 @@ GdipCloneRegion (GpRegion *region, GpRegion **cloneRegion)
                 return InvalidParameter;
 
         result = (GpRegion *) GdipAlloc (sizeof (GpRegion));
-        memcpy (result, region, sizeof (GpRegion));
-        result->rects = (GpRectF *) GdipAlloc (sizeof (GpRectF) * region->cnt);
-        memcpy (result->rects, region->rects, sizeof (GpRectF) * region->cnt);
+	gdip_copy_region (region, result);
         *cloneRegion = result;
 
         return Ok;
@@ -281,9 +466,7 @@ GdipDeleteRegion (GpRegion *region)
         if (!region)
                 return InvalidParameter;
 
-        if (region->rects)
-                GdipFree (region->rects);
-
+	gdip_clear_region (region);
         GdipFree (region);
 
         return Ok;
@@ -298,10 +481,11 @@ GdipSetInfinite (GpRegion *region)
         if (!region)
                 return InvalidParameter;
 
-        GdipSetEmpty (region);
+	gdip_clear_region (region);
+	region->type = RegionTypeRectF;
 
-        rect.X = rect.Y = -4194304;
-        rect.Width = rect.Height= 8388608;
+        rect.X = rect.Y = REGION_INFINITE_POSITION;
+        rect.Width = rect.Height = REGION_INFINITE_LENGTH;
 
         gdip_add_rect_to_array (&region->rects, &region->cnt,  &rect);
         return Ok;
@@ -314,11 +498,8 @@ GdipSetEmpty (GpRegion *region)
         if (!region)
                 return InvalidParameter;
 
-        if (region->rects)
-                GdipFree (region->rects);
-
-        region->rects = NULL;
-        region->cnt = 0;
+	gdip_clear_region (region);
+	region->type = RegionTypeRectF;
 
         return Ok;
 }
@@ -359,8 +540,6 @@ gdip_get_bounds (GpRectF *allrects, int allcnt, GpRectF *bound)
 }
 
 
-
-
 /* Exclude */
 void
 gdip_combine_exclude (GpRegion *region, GpRectF *rtrg, int cntt)
@@ -462,7 +641,8 @@ gdip_combine_exclude (GpRegion *region, GpRectF *rtrg, int cntt)
 			break;
 		}
 
-		if (storecomplete) {
+		/* don't include a rectangle identical to the excluded one! */
+		if (storecomplete && !gdip_equals (rtrg, &current)) {
 			gdip_add_rect_to_array_notcontained (&rects, &cnt,  &current);
 		}
 	}
@@ -612,6 +792,7 @@ gdip_combine_union (GpRegion *region, GpRectF *rtrg, int cnttrg)
 
 			if (contained == FALSE && rslt.Height > 0 && rslt.Width > 0) {
 				gdip_add_rect_to_array (&allrects, &allcnt,  &rslt);
+				recttrg = allrects;
 			}
 
 			/* If both we at the same Y when take into account the X also to process the following
@@ -714,12 +895,15 @@ gdip_combine_xor (GpRegion *region, GpRectF *recttrg, int cnttrg)
                 gdip_add_rect_to_array (&allrects, &allcnt,  rect);
 
 	rgnsrc = (GpRegion *) GdipAlloc (sizeof (GpRegion));
+	rgnsrc->type = RegionTypeRectF;
         rgnsrc->cnt = allcnt;
         rgnsrc->rects = allrects;
 
         GdipCloneRegion (region, &rgntrg);
         gdip_combine_intersect (rgntrg, recttrg, cnttrg);
-        gdip_combine_exclude (rgnsrc, rgntrg->rects, rgntrg->cnt);
+	/* exclude the intersecting rectangles (if any) */
+	if (rgntrg->cnt > 0)
+		gdip_combine_exclude (rgnsrc, rgntrg->rects, rgntrg->cnt);
 
         if (region->rects)
                 GdipFree (region->rects);
@@ -738,6 +922,19 @@ GdipCombineRegionRect (GpRegion *region, GDIPCONST GpRectF *rect, CombineMode co
         if (!region || !rect)
                 return InvalidParameter;
 
+	if (region->type == RegionTypePath) {
+		GpPath *path = NULL;
+		GpStatus status;
+
+		/* Convert GpRectF to GpPath and use GdipCombineRegionPath */
+		GdipCreatePath (FillModeAlternate, &path);
+		GdipAddPathRectangle (path, rect->X, rect->Y, rect->Width, rect->Height);
+		status = GdipCombineRegionPath (region, path, combineMode);
+		GdipDeletePath (path);
+		return status;
+	}
+
+	/* region is rectangle-based */
         switch (combineMode) {
         case CombineModeExclude:
                 gdip_combine_exclude (region, (GpRectF *) rect, 1);
@@ -757,9 +954,9 @@ GdipCombineRegionRect (GpRegion *region, GDIPCONST GpRectF *rect, CombineMode co
 	case CombineModeReplace: /* Used by Graphics clipping */
 		GdipSetEmpty (region);
 		gdip_add_rect_to_array (&region->rects, &region->cnt, (GpRectF *)rect);
+		break;
         default:
                return NotImplemented;
-
         }
 
         return Ok;
@@ -783,7 +980,136 @@ GdipCombineRegionRectI (GpRegion *region, GDIPCONST GpRect *recti, CombineMode c
 GpStatus
 GdipCombineRegionPath (GpRegion *region, GpPath *path, CombineMode combineMode)
 {
-        return NotImplemented;
+	GpRegionBitmap *path_bitmap, *result;
+	GpPathTree* tmp;
+
+	if (!region || !path)
+		return InvalidParameter;
+
+	/* special case #1 - replace */
+	if (combineMode == CombineModeReplace) {
+		gdip_clear_region (region);
+		gdip_region_create_from_path (region, path);
+		return Ok;
+	}
+
+	/* special case #2 - the region is empty */
+	if (gdip_is_region_empty (region)) {
+		switch (combineMode) {
+		case CombineModeComplement:
+		case CombineModeUnion:
+		case CombineModeXor:
+			/* this is like "adding" the path */
+			gdip_clear_region (region);
+			gdip_region_create_from_path (region, path);
+			break;
+		default:
+			/* Intersect and Exclude are no-op on an empty region */
+			break;
+		}
+		return Ok;
+	}
+
+	/* special case #3 - the region is infinite */
+	if (gdip_is_InfiniteRegion (region)) {
+		switch (combineMode) {
+		case CombineModeUnion:
+			/* Union with infinity is a no-op (still an infinite region) */
+			return Ok;
+		case CombineModeComplement:
+			/* Complement of infinity is empty - whatever the path is */
+			gdip_clear_region (region);
+			region->type = RegionTypeRectF;
+			return Ok;
+		case CombineModeIntersect:
+			/* Intersection with infinity is the path itself */
+			gdip_clear_region (region);
+			gdip_region_create_from_path (region, path);
+			return Ok;
+		default:
+			/* Xor and Exclude must be treated as a "normal" case */
+			break;
+		}
+	}
+
+	if (region->type == RegionTypeRectF)
+		gdip_region_convert_to_path (region);
+
+	/* make sure the region's bitmap is available */
+	gdip_region_bitmap_ensure (region);
+	g_assert (region->bitmap);
+
+	/* create a bitmap for the path to combine into the region */
+	path_bitmap = gdip_region_bitmap_from_path (path);
+
+	result = gdip_region_bitmap_combine (region->bitmap, path_bitmap, combineMode);
+	gdip_region_bitmap_free (path_bitmap);
+	if (!result)
+		return NotImplemented;
+
+	gdip_region_bitmap_free (region->bitmap);
+	region->bitmap = result;
+
+	/* add a copy of path into region1 tree */
+	tmp = (GpPathTree*) GdipAlloc (sizeof (GpPathTree));
+	tmp->mode = combineMode;
+	tmp->path = NULL;
+	tmp->branch1 = (GpPathTree*) GdipAlloc (sizeof (GpPathTree));
+	tmp->branch2 = (GpPathTree*) GdipAlloc (sizeof (GpPathTree));
+
+	if (region->tree->path) {
+		tmp->branch1->path = region->tree->path;
+	} else {
+		gdip_region_copy_tree (region->tree, tmp->branch1);
+	}
+
+	GdipClonePath (path, &tmp->branch2->path);
+
+	region->tree = tmp;
+	return Ok;
+}
+
+
+static GpStatus
+gdip_combine_pathbased_region (GpRegion *region1, GpRegion *region2, CombineMode combineMode)
+{
+	GpRegionBitmap *result;
+	GpPathTree* tmp;
+
+	/* if not available, construct the bitmaps for both regions */
+	gdip_region_bitmap_ensure (region1);
+	gdip_region_bitmap_ensure (region2);
+	if (!region1->bitmap || !region2->bitmap)
+		return OutOfMemory;
+
+	result = gdip_region_bitmap_combine (region1->bitmap, region2->bitmap, combineMode);
+	if (!result)
+		return NotImplemented;
+	gdip_region_bitmap_free (region1->bitmap);
+	region1->bitmap = result;
+
+	/* add a copy of region2 tree into region1 tree */
+	tmp = (GpPathTree*) GdipAlloc (sizeof (GpPathTree));
+	tmp->mode = combineMode;
+	tmp->path = NULL;
+	tmp->branch1 = (GpPathTree*) GdipAlloc (sizeof (GpPathTree));
+	tmp->branch2 = (GpPathTree*) GdipAlloc (sizeof (GpPathTree));
+
+	if (region1->tree->path) {
+		tmp->branch1->path = region1->tree->path;
+	} else {
+		gdip_region_copy_tree (region1->tree, tmp->branch1);
+	}
+
+	if (region2->tree->path) {
+		GdipClonePath (region2->tree->path, &tmp->branch2->path);
+	} else {
+		gdip_region_copy_tree (region2->tree, tmp->branch2);
+	}
+
+	GdipFree (region1->tree);
+	region1->tree = tmp;
+	return Ok;
 }
 
 
@@ -793,6 +1119,25 @@ GdipCombineRegionRegion (GpRegion *region,  GpRegion *region2, CombineMode combi
         if (!region || !region2)
                 return InvalidParameter;
 
+	/* special case to deal with copying empty and infinity regions */
+	/* CombineModeReplace is used by Graphics clipping */
+	if (combineMode == CombineModeReplace) {
+		GdipSetEmpty (region);
+		gdip_copy_region (region2, region);
+		return Ok;
+	}
+
+	if (region->type == RegionTypePath) {
+		gdip_region_convert_to_path (region2);
+		return gdip_combine_pathbased_region (region, region2, combineMode);
+	} else if (region2->type == RegionTypePath) {
+		gdip_region_convert_to_path (region);
+		return gdip_combine_pathbased_region (region, region2, combineMode);
+	}
+
+	/* at this stage we are sure that BOTH region and region2 are rectangle 
+	 * based, so we can use the old rectangle-based code to combine regions
+	 */
         switch (combineMode) {
         case CombineModeExclude:
                 gdip_combine_exclude (region, region2->rects, region2->cnt);
@@ -809,19 +1154,8 @@ GdipCombineRegionRegion (GpRegion *region,  GpRegion *region2, CombineMode combi
         case CombineModeXor:
                 gdip_combine_xor (region, region2->rects, region2->cnt);
                 break;
-	case CombineModeReplace: { /* Used by Graphics clipping */
-		int i;
-		GpRectF* rect;
-		GdipSetEmpty (region);
-		
-		for (i = 0, rect = region2->rects; i < region2->cnt; i++, rect++)
-			gdip_add_rect_to_array (&region->rects, &region->cnt, rect);
-	
-		break;
-	}
         default:
                return NotImplemented;
-
         }
 
         return Ok;
@@ -833,7 +1167,24 @@ GdipGetRegionBounds (GpRegion *region, GpGraphics *graphics, GpRectF *rect)
         if (!region || !graphics || !rect)
                 return InvalidParameter;
 
-        gdip_get_bounds (region->rects , region->cnt, rect);
+	if (region->type == RegionTypePath) {
+		GpRect bounds;
+
+		gdip_region_bitmap_ensure (region);
+		if (!region->bitmap)
+			return OutOfMemory;
+
+		/* base the bounds on the reduced bitmap of the paths */
+		gdip_region_bitmap_get_smallest_rect (region->bitmap, &bounds);
+
+		/* small loss of precision when converting the bitmap coord to float */
+		rect->X = bounds.X;
+		rect->Y = bounds.Y;
+		rect->Width = bounds.Width;
+		rect->Height = bounds.Height;
+	} else {
+	        gdip_get_bounds (region->rects , region->cnt, rect);
+	}
 
         return Ok;
 }
@@ -842,18 +1193,10 @@ GdipGetRegionBounds (GpRegion *region, GpGraphics *graphics, GpRectF *rect)
 GpStatus
 GdipIsEmptyRegion (GpRegion *region, GpGraphics *graphics, BOOL *result)
 {
-        GpRectF rect;
-
         if (!region || !graphics || !result)
                 return InvalidParameter;
 
-        gdip_get_bounds (region->rects, region->cnt, &rect);
-
-        if (rect.Width ==0  || rect.Height ==0)
-                *result = TRUE;
-        else
-                *result = FALSE;
-
+	*result = gdip_is_region_empty (region);
         return Ok;
 }
 
@@ -872,11 +1215,17 @@ GdipIsInfiniteRegion (GpRegion *region, GpGraphics *graphics, BOOL *result)
 GpStatus
 GdipIsVisibleRegionPoint (GpRegion *region, float x, float y, GpGraphics *graphics, BOOL *result)
 {
-
         if (!region || !result)
                 return InvalidParameter;
 
-        *result = gdip_is_Point_in_RectFs_Visible (x, y, region->rects, region->cnt);
+	if (region->type == RegionTypePath) {
+		gdip_region_bitmap_ensure (region);
+		g_assert (region->bitmap);
+
+		*result = gdip_region_bitmap_is_point_visible (region->bitmap, x, y);
+	} else {
+		*result = gdip_is_Point_in_RectFs_Visible (x, y, region->rects, region->cnt);
+	}
 
         return Ok;
 }
@@ -885,15 +1234,7 @@ GdipIsVisibleRegionPoint (GpRegion *region, float x, float y, GpGraphics *graphi
 GpStatus
 GdipIsVisibleRegionPointI (GpRegion *region, int x, int y, GpGraphics *graphics, BOOL *result)
 {
-        float xf, yf;
-
-        if (!region || !result)
-                return InvalidParameter;
-
-        xf = x;
-        yf = y;
-
-        return GdipIsVisibleRegionPoint (region, xf, yf, graphics, result);
+        return GdipIsVisibleRegionPoint (region, x, y, graphics, result);
 }
 
 
@@ -901,8 +1242,6 @@ GpStatus
 GdipIsVisibleRegionRect (GpRegion *region, float x, float y, float width, float height, GpGraphics *graphics, BOOL *result)
 {
         BOOL found = FALSE;
-        float posy, posx;
-        GpRectF recthit;
 
         if (!region || !result)
                 return InvalidParameter;
@@ -912,19 +1251,37 @@ GdipIsVisibleRegionRect (GpRegion *region, float x, float y, float width, float
                 return Ok;
         }
 
-        recthit.X = x; recthit.Y = y;
-        recthit.Width = width; recthit.Height = height;
+	if (region->type == RegionTypePath) {
+		GpRect rect;
 
-        /* Any point of intersection ?*/
-        for (posy = 0; posy < recthit.Height && found == FALSE; posy++) {
+		rect.X = x;
+		rect.Y = y;
+		rect.Width = width;
+		rect.Height = height;
+		
+		gdip_region_bitmap_ensure (region);
+		g_assert (region->bitmap);
 
-                for (posx = 0; posx < recthit.Width ; posx++) {
-                        if (gdip_is_Point_in_RectFs_Visible (recthit.X + posx , recthit.Y + posy, region->rects, region->cnt) == TRUE) {
-                                found = TRUE;
-                                break;
-                        }
+		found = gdip_region_bitmap_is_rect_visible (region->bitmap, &rect);
+	} else {
+		float posy, posx;
+		GpRectF recthit;
+
+		recthit.X = x;
+		recthit.Y = y;
+		recthit.Width = width;
+		recthit.Height = height;
+
+		/* Any point of intersection ?*/
+		for (posy = 0; posy < recthit.Height && found == FALSE; posy++) {
+			for (posx = 0; posx < recthit.Width ; posx++) {
+				if (gdip_is_Point_in_RectFs_Visible (recthit.X + posx , recthit.Y + posy, region->rects, region->cnt) == TRUE) {
+					found = TRUE;
+					break;
+				}
+			}
 		}
-        }
+	}
 
         *result = found;
         return Ok;
@@ -934,37 +1291,115 @@ GdipIsVisibleRegionRect (GpRegion *region, float x, float y, float width, float
 GpStatus
 GdipIsVisibleRegionRectI (GpRegion *region, int x, int y, int width, int height, GpGraphics *graphics, BOOL *result)
 {
-        float fx, fy, fw, fh;
-
-        if (!region || !result)
-                return InvalidParameter;
-
-        fx = x; fy = y;
-        fh = height; fw = width;
-
-        return GdipIsVisibleRegionRect (region, fx, fy, fw, fh, graphics, result);
+        return GdipIsVisibleRegionRect (region, x, y, width, height, graphics, result);
 }
 
 
 GpStatus
 GdipGetRegionScansCount (GpRegion *region, int* count, GpMatrix* matrix)
 {
+	GpRegion *work;
+	GpStatus status;
+
         if (!region || !count)
                 return InvalidParameter;
 
-        *count = region->cnt;
-        return Ok;
+	/* apply any user supplied matrix transformation */
+	if (!gdip_is_matrix_empty (matrix)) {
+		int i;
+
+		/* the matrix doesn't affect the original region - only the result */
+		status = GdipCloneRegion (region, &work);
+		if (status != Ok)
+			return status;
+
+		/* if required convert into a path-based region */
+		if (work->type != RegionTypePath)
+			gdip_region_convert_to_path (work);
+
+		/* transform all the paths */
+		status = gdip_region_transform_tree (work->tree, matrix);
+		if (status != Ok) {
+			GdipDeleteRegion (work);
+			return status;
+		}
+		/* note: any existing bitmap has been invalidated */
+		gdip_region_bitmap_invalidate (work);
+	} else {
+		work = region;
+	}
+
+	if (work->type == RegionTypePath) {
+		/* ensure the bitmap is usable */
+		gdip_region_bitmap_ensure (work);
+
+		/* check if region is too large to render */
+		if (work->bitmap)
+			*count = gdip_region_bitmap_get_scans (work->bitmap, NULL, -1);
+		else
+			*count = 0;
+	} else {
+	        *count = work->cnt;
+	}
+
+	/* delete the clone */
+	if (work != region)
+		GdipDeleteRegion (work);
+	return Ok;
 }
 
 GpStatus
 GdipGetRegionScans (GpRegion *region, GpRectF* rects, int* count, GpMatrix* matrix)
 {
+	GpRegion *work;
+	GpStatus status;
+
         if (!region || !rects|| !count)
                 return InvalidParameter;
 
-        memcpy (rects, region->rects, sizeof (GpRectF) * *count);
-        *count = region->cnt;
-        return Ok;
+	/* apply any user supplied matrix transformation */
+	if (!gdip_is_matrix_empty (matrix)) {
+		int i;
+
+		/* the matrix doesn't affect the original region - only the result */
+		status = GdipCloneRegion (region, &work);
+		if (status != Ok)
+			return status;
+
+		/* if required convert into a path-based region */
+		if (work->type != RegionTypePath)
+			gdip_region_convert_to_path (work);
+
+		/* transform all the paths */
+		status = gdip_region_transform_tree (work->tree, matrix);
+		if (status != Ok) {
+			GdipDeleteRegion (work);
+			return status;
+		}
+		/* note: any existing bitmap has been invalidated */
+		gdip_region_bitmap_invalidate (work);
+	} else {
+		work = region;
+	}
+
+	if (region->type == RegionTypePath) {
+		/* ensure the bitmap is usable */
+		gdip_region_bitmap_ensure (work);
+
+		/* check if region is too large to render */
+		if (work->bitmap)
+			*count = gdip_region_bitmap_get_scans (work->bitmap, rects, *count);
+		else
+			*count = 0;
+	} else {
+	        memcpy (rects, work->rects, sizeof (GpRectF) * *count);
+	        *count = work->cnt;
+	}
+
+	/* delete the clone */
+	if (work != region)
+		GdipDeleteRegion (work);
+	return Ok;
 }
 
 GpStatus
@@ -976,6 +1411,30 @@ GdipIsEqualRegion (GpRegion *region, GpRegion *region2, GpGraphics *graphics, BO
         if (!region || !region2 || !graphics || !result)
                 return InvalidParameter;
 
+	/* quick case: same pointer == same region == equals */
+	if (region == region2) {
+		*result = TRUE;
+		return Ok;
+	}
+
+	if ((region->type == RegionTypePath) || (region2->type == RegionTypePath)) {
+		/* if required convert one region to a path based region */
+		if (region->type != RegionTypePath)
+			gdip_region_convert_to_path (region);
+		gdip_region_bitmap_ensure (region);
+		g_assert (region->bitmap);
+
+		if (region2->type != RegionTypePath)
+			gdip_region_convert_to_path (region2);
+
+		gdip_region_bitmap_ensure (region2);
+		g_assert (region2->bitmap);
+
+		*result = gdip_region_bitmap_compare (region->bitmap, region2->bitmap);
+                return Ok;
+	}
+
+	/* rectangular-based region quality test */
         if (region->cnt != region2->cnt) {
                 *result = FALSE;
                 return Ok;
@@ -998,15 +1457,20 @@ GdipIsEqualRegion (GpRegion *region, GpRegion *region2, GpGraphics *graphics, BO
 GpStatus
 GdipTranslateRegion (GpRegion *region, float dx, float dy)
 {
-        int i;
-        GpRectF *rect;
-
         if (!region)
                 return InvalidParameter;
 
-        if (region->rects) {
-
-                for (i = 0, rect=region->rects ; i < region->cnt; i++, rect++) {
+	if (region->type == RegionTypePath) {
+		gdip_region_translate_tree (region->tree, dx, dy);
+		/* any existing bitmap is still valid _if_ we update it's origin */
+		if (region->bitmap) {
+			region->bitmap->X += dx;
+			region->bitmap->Y += dy;
+		}
+	} else if ((region->type == RegionTypeRectF) && region->rects) {
+	        int i;
+	        GpRectF *rect;
+                for (i = 0, rect = region->rects ; i < region->cnt; i++, rect++) {
                         rect->X += dx;
                         rect->Y += dy;
                 }
@@ -1018,39 +1482,154 @@ GdipTranslateRegion (GpRegion *region, float dx, float dy)
 GpStatus
 GdipTranslateRegionI (GpRegion *region, int dx, int dy)
 {
-        float fx, fy;
-
-        if (!region)
-                return InvalidParameter;
-
-        fx = dx; fy = dy;
-        return  GdipTranslateRegion (region, fx, fy);
+        return GdipTranslateRegion (region, dx, dy);
 }
 
 GpStatus
 GdipTransformRegion (GpRegion *region, GpMatrix *matrix)
 {
-        return NotImplemented;
+	GpStatus status;
+
+	if (!region || !matrix)
+		return InvalidParameter;
+
+	/* don't (possibly) convert to a bitmap if the matrix is empty (a no-op) */
+	if (gdip_is_matrix_empty (matrix))
+		return Ok;
+
+	/* avoid heavy stuff (e.g. convertion to path, invalidating bitmap...)
+	   if the transform is only to do a simple translation */
+	if (gdip_is_matrix_a_translation (matrix)) {
+		return GdipTranslateRegion (region, 
+			gdip_matrix_get_x_translation (matrix), 
+			gdip_matrix_get_y_translation (matrix));
+	}
+
+	/* most matrix operations would change the rectangles into path so we always preempt this */
+	if (region->type != RegionTypePath)
+		gdip_region_convert_to_path (region);
+
+	/* apply the same transformation matrix to all paths */
+	status = gdip_region_transform_tree (region->tree, matrix);
+
+	/* invalidate the bitmap so it will get re-created on the next gdip_region_bitmap_ensure call */
+	gdip_region_bitmap_invalidate (region);
+
+        return status;
 }
 
 GpStatus
 GdipCreateRegionPath (GpPath *path, GpRegion **region)
 {
-        return NotImplemented;
+        if (!region || !path)
+                return InvalidParameter;
+
+        return gdip_createRegion (region, RegionTypePath, (void*) path);
+}
+
+
+/*
+ * The internal data representation for RegionData depends on the type of region.
+ *
+ * Type 0 (RegionTypeEmpty)
+ *	Note: There is no type 0. RegionTypeEmpty are converted into 
+ *	RegionTypeRectF (type 2) when a region is created.
+ *
+ * Type 1 (RegionTypeRect)
+ *	Note: There is no type 1. RegionTypeRect are converted into 
+ *	RegionTypeRectF (type 2) when a region is created.
+ *
+ * Type 2 (RegionTypeRectF), variable size
+ *	guint32 RegionType	Always 2
+ *	guint32 Count		0-2^32
+ *	GpRectF[Count] Points
+ *
+ * Type 3 (RegionTypePath), variable size
+ *	guint32 RegionType	Always 3
+ *	GpPathTree tree
+ *
+ * where GpPathTree is
+ *	guint32 Tag		1 = Path, 2 = Tree
+ *	data[n]
+ *
+ *	where data is for tag 1 (Path)
+ *		guint32 Count		0-2^32
+ *		GpFillMode FillMode	
+ *		guint8[Count] Types	
+ *		GpPointF[Count] Points
+ *	or
+ *	where data is for tag 2 (Tree)
+ *		guint32 Operation	see CombineMode
+ *		guint32 Size1		0-2^32
+ *		byte[Size1]		branch #1
+ *		guint32 Size2		0-2^32
+ *		byte[Size2]		branch #2
+ */
+
+GpStatus
+GdipGetRegionDataSize (GpRegion *region, UINT *bufferSize)
+{
+	if (!region || !bufferSize)
+		return InvalidParameter;
+
+	switch (region->type) {
+	case RegionTypeRectF:
+		*bufferSize = (sizeof (guint32) * 2) + (region->cnt * sizeof (GpRectF));
+		break;
+	case RegionTypePath:
+		/* regiontype, tree */
+		*bufferSize = sizeof (guint32) + gdip_region_get_tree_size (region->tree);
+		break;
+	default:
+		g_warning ("unknown type %d", region->type);
+		return NotImplemented;
+	}
+	return Ok;
 }
 
 
 GpStatus
-GdipGetRegionDataSize (GpRegion *region, UINT * bufferSize)
+GdipGetRegionData (GpRegion *region, BYTE *buffer, UINT bufferSize, UINT *sizeFilled)
 {
-        return NotImplemented;
+	GpStatus status;
+	UINT size;
+	int len;
+
+	if (!region || !buffer || !sizeFilled)
+		return InvalidParameter;
+
+	status = GdipGetRegionDataSize (region, &size);
+	if (status != Ok)
+		return status;
+	if (size > bufferSize)
+		return InsufficientBuffer;
+
+	/* type of region */
+	len = sizeof (guint32);
+	memcpy (buffer, &region->type, len);
+	buffer += len;
+	*sizeFilled += len;
+
+	switch (region->type) {
+	case RegionTypeRectF:
+		/* count (# rectangles) */
+		memcpy (buffer, &region->cnt, len);
+		buffer += len;
+		*sizeFilled += len;
+		/* rectangles */
+		len = sizeof (GpRectF) * (region->cnt);
+		memcpy (buffer, region->rects, len);
+		*sizeFilled += len;
+		break;
+	case RegionTypePath:
+		bufferSize -= len;
+		if (!gdip_region_serialize_tree (region->tree, buffer, bufferSize, sizeFilled))
+			return InsufficientBuffer;
+		break;
+	default:
+		g_warning ("unknown type %d", region->type);
+		return NotImplemented;
+	}
+
+	return Ok;
 }
-
-
-GpStatus
-GdipGetRegionData(GpRegion *region, BYTE * buffer, UINT bufferSize, UINT *sizeFilled)
-{
-        return NotImplemented;
-}
-
-
