@@ -24,6 +24,7 @@
  */
 
 #include "gdip.h"
+#include "gdipImage.h"
 #include "texturebrush.h"
 
 static GpStatus gdip_texture_setup (GpGraphics *graphics, GpBrush *brush);
@@ -71,6 +72,7 @@ draw_tile_texture (cairo_t *ct, GpBitmap *bitmap, GpTexture *brush)
 	cairo_surface_t *texture;
 	cairo_pattern_t *pat, *pattern;
 	GpRect *rect = brush->rectangle;
+
 	g_return_val_if_fail (rect != NULL, InvalidParameter);
 
 	/* Original image surface */
@@ -85,22 +87,22 @@ draw_tile_texture (cairo_t *ct, GpBitmap *bitmap, GpTexture *brush)
 	/* texture surface to be created */
 	texture = cairo_surface_create_similar (original, from_cairoformat_to_content (bitmap->cairo_format),
 						2 * rect->Width, 2 * rect->Height);
+	if (texture == NULL) {
+		cairo_pattern_destroy (pat);
+		return OutOfMemory;
+	} else {
+		cairo_t *ct2;
 
-	g_return_val_if_fail (texture != NULL, OutOfMemory);
-
-	cairo_save (ct);
-	{
 		/* Draw the texture */
-		ct = cairo_create (texture);
-		cairo_identity_matrix (ct);
-		cairo_set_source (ct, pat);
-		cairo_rectangle (ct, 0, 0, 2 * rect->Width, 2 * rect->Height);
-		cairo_fill (ct);
+		ct2 = cairo_create (texture);
+		cairo_identity_matrix (ct2);
+		cairo_set_source (ct2, pat);
+		cairo_rectangle (ct2, 0, 0, 2 * rect->Width, 2 * rect->Height);
+		cairo_fill (ct2);
+		cairo_destroy(ct2);
 	}
-	cairo_restore (ct);
-	pattern = cairo_pattern_create_for_surface (texture);
-	cairo_pattern_set_extend (pattern, CAIRO_EXTEND_REPEAT);
 	brush->pattern = cairo_pattern_create_for_surface (texture);
+	cairo_pattern_set_extend (brush->pattern, CAIRO_EXTEND_REPEAT);
 
 	cairo_pattern_destroy (pat);
 	cairo_surface_destroy (texture);
@@ -114,7 +116,6 @@ draw_tile_flipX_texture (cairo_t *ct, GpBitmap *bitmap, GpTexture *brush)
 	cairo_surface_t *original;
 	cairo_surface_t *texture;
 	cairo_pattern_t *pat, *pattern;
-	cairo_matrix_t tempMatrix;
 	GpRect *rect = brush->rectangle;
 	g_return_val_if_fail (rect != NULL, InvalidParameter);
 
@@ -129,40 +130,41 @@ draw_tile_flipX_texture (cairo_t *ct, GpBitmap *bitmap, GpTexture *brush)
 	/* texture surface to be created */
 	texture = cairo_surface_create_similar (original, from_cairoformat_to_content (bitmap->cairo_format),
 						2 * rect->Width, rect->Height);
-
 	if (texture == NULL) {
+		cairo_pattern_destroy (pat);
 		return OutOfMemory;
-	}
+	} else {
+		cairo_t	*ct2;
+		cairo_matrix_t tempMatrix;
 
-	cairo_save (ct);
-	{
-		cairo_identity_matrix (ct);
 		/* Draw left part of the texture */
-		ct = cairo_create (texture);
-		cairo_set_source (ct, pat);
-		cairo_rectangle (ct, 0, 0, rect->Width, rect->Height);
-		cairo_fill (ct);
+		ct2 = cairo_create (texture);
+		cairo_set_source (ct2, pat);
+		cairo_rectangle (ct2, 0, 0, rect->Width, rect->Height);
+		cairo_fill (ct2);
 		
-		/* Draw right part of the texture */
-		cairo_translate (ct, rect->Width, 0);
-		cairo_rectangle (ct, 0, 0, rect->Width, rect->Height);
-
 		/* Not sure if this is a bug, but using rect->Width - 1
 		 * avoids the seam.
 		 */
+		cairo_matrix_init_identity (&tempMatrix);
 		cairo_matrix_translate (&tempMatrix, rect->Width - 1, 0);
 		/* scale in -X direction to flip along X */
 		cairo_matrix_scale (&tempMatrix, -1.0, 1.0);
 		cairo_pattern_set_matrix (pat, &tempMatrix);
-		cairo_fill (ct);
+
+		/* Draw right part of the texture */
+		cairo_translate (ct2, rect->Width, 0);
+		cairo_set_source (ct2, pat);
+		cairo_rectangle (ct2, 0, 0, rect->Width, rect->Height);
+
+		cairo_fill (ct2);
+		cairo_destroy(ct2);
 	}
-	cairo_restore (ct);
 
-	pattern = cairo_pattern_create_for_surface (texture);
-	cairo_pattern_set_extend (pattern, CAIRO_EXTEND_REPEAT);
-	
 	brush->pattern = cairo_pattern_create_for_surface (texture);
+	cairo_pattern_set_extend (brush->pattern, CAIRO_EXTEND_REPEAT);
 
+	cairo_pattern_destroy (pat);
 	cairo_surface_destroy (texture);
 
 	return gdip_get_status (cairo_status (ct));
@@ -189,39 +191,39 @@ draw_tile_flipY_texture (cairo_t *ct, GpBitmap *bitmap, GpTexture *brush)
 	/* texture surface to be created */
 	texture = cairo_surface_create_similar (original, from_cairoformat_to_content (bitmap->cairo_format),
 						rect->Width, 2 * rect->Height);
-
 	if (texture == NULL) {
+		cairo_pattern_destroy (pat);
 		return OutOfMemory;
-	}
+	} else {
+		cairo_t	*ct2;
 
-	cairo_save (ct);
-	{
-		cairo_identity_matrix (ct);
 		/* Draw upper part of the texture */
-		ct = cairo_create (texture);
-		cairo_set_source (ct, pat);
-		cairo_rectangle (ct, 0, 0, rect->Width, rect->Height);
-		cairo_fill (ct);
+		ct2 = cairo_create (texture);
+		cairo_set_source (ct2, pat);
+		cairo_rectangle (ct2, 0, 0, rect->Width, rect->Height);
+		cairo_fill (ct2);
 		
-		/* Draw lower part of the texture */
-		cairo_translate (ct, 0, rect->Height);
-		cairo_rectangle (ct, 0, 0, rect->Width, rect->Height);
-
 		/* Not sure if this is a bug, but using rect->Height - 1
 		 * avoids the seam.
 		 */
+		cairo_matrix_init_identity (&tempMatrix);
 		cairo_matrix_translate (&tempMatrix, 0, rect->Height - 1);
 		/* scale in -Y direction to flip along Y */
 		cairo_matrix_scale (&tempMatrix, 1.0, -1.0);
 		cairo_pattern_set_matrix (pat, &tempMatrix);
-		cairo_fill (ct);
+
+		/* Draw lower part of the texture */
+		cairo_translate (ct2, 0, rect->Height);
+		cairo_set_source (ct2, pat);
+		cairo_rectangle (ct2, 0, 0, rect->Width, rect->Height);
+		cairo_fill (ct2);
+		cairo_destroy(ct2);
 	}
-	cairo_restore (ct);
 
-	pattern = cairo_pattern_create_for_surface (texture);
-	cairo_pattern_set_extend (pattern, CAIRO_EXTEND_REPEAT);	
 	brush->pattern = cairo_pattern_create_for_surface (texture);
+	cairo_pattern_set_extend (brush->pattern, CAIRO_EXTEND_REPEAT);	
 
+	cairo_pattern_destroy (pat);
 	cairo_surface_destroy (texture);
 
 	return gdip_get_status (cairo_status (ct));
@@ -248,36 +250,34 @@ draw_tile_flipXY_texture (cairo_t *ct, GpBitmap *bitmap, GpTexture *brush)
 	/* texture surface to be created */
 	texture = cairo_surface_create_similar (original, from_cairoformat_to_content (bitmap->cairo_format),
 						2 * rect->Width, 2 * rect->Height);
-
 	if (texture == NULL) {
+		cairo_pattern_destroy (pat);
 		return OutOfMemory;
-	}
+	} else {
+		cairo_t *ct2;
 
-	cairo_save (ct);
-	{
-		cairo_identity_matrix (ct);
-		ct = cairo_create (texture);
-		cairo_set_source (ct, pat);
+		ct2 = cairo_create (texture);
+		cairo_set_source (ct2, pat);
 
 		/* Draw upper left part of the texture */
-		cairo_rectangle (ct, 0, 0, rect->Width, rect->Height);
-		cairo_fill (ct);
+		cairo_rectangle (ct2, 0, 0, rect->Width, rect->Height);
+		cairo_fill (ct2);
 
-		/* Draw lower left part of the texture */
-		cairo_translate (ct, 0, rect->Height);
-		cairo_rectangle (ct, 0, 0, rect->Width, rect->Height);
 		/* Not sure if this is a bug, but using rect->Height - 1
 		 * avoids the seam.
 		 */
+		cairo_matrix_init_identity (&tempMatrix);
 		cairo_matrix_translate (&tempMatrix, 0, rect->Height - 1);
 		/* scale in -Y direction to flip along Y */
 		cairo_matrix_scale (&tempMatrix, 1.0, -1.0);
 		cairo_pattern_set_matrix (pat, &tempMatrix);
-		cairo_fill (ct);
 
-		/* Draw upper right part of the texture */
-		cairo_translate (ct, rect->Width, -rect->Height);
-		cairo_rectangle (ct, 0, 0, rect->Width, rect->Height);
+		/* Draw lower left part of the texture */
+		cairo_translate (ct2, 0, rect->Height);
+		cairo_set_source (ct2, pat);
+		cairo_rectangle (ct2, 0, 0, rect->Width, rect->Height);
+		cairo_fill (ct2);
+
 		/* Reset the pattern matrix and do fresh transformation */
 		cairo_matrix_init_identity (&tempMatrix);
 		/* Not sure if this is a bug, but using rect->Width - 1
@@ -287,11 +287,13 @@ draw_tile_flipXY_texture (cairo_t *ct, GpBitmap *bitmap, GpTexture *brush)
 		/* scale in -X direction to flip along X */
 		cairo_matrix_scale (&tempMatrix, -1.0, 1.0);
 		cairo_pattern_set_matrix (pat, &tempMatrix);
-		cairo_fill (ct);
 
-		/* Draw lower right part of the texture */
-		cairo_translate (ct, 0, rect->Height);
-		cairo_rectangle (ct, 0, 0, rect->Width, rect->Height);
+		/* Draw upper right part of the texture */
+		cairo_translate (ct2, rect->Width, -rect->Height);
+		cairo_set_source (ct2, pat);
+		cairo_rectangle (ct2, 0, 0, rect->Width, rect->Height);
+		cairo_fill (ct2);
+
 		/* Not sure if this is a bug, but using rect->Height - 1
 		 * avoids the seam.
 		 */
@@ -299,15 +301,19 @@ draw_tile_flipXY_texture (cairo_t *ct, GpBitmap *bitmap, GpTexture *brush)
 		/* scale in -Y direction to flip along Y */
 		cairo_matrix_scale (&tempMatrix, 1.0, -1.0);
 		cairo_pattern_set_matrix (pat, &tempMatrix);
-		cairo_fill (ct);
+
+		/* Draw lower right part of the texture */
+		cairo_translate (ct2, 0, rect->Height);
+		cairo_set_source (ct2, pat);
+		cairo_rectangle (ct2, 0, 0, rect->Width, rect->Height);
+		cairo_fill (ct2);
+		cairo_destroy(ct2);
 	}
-	cairo_restore (ct);
 
-	pattern = cairo_pattern_create_for_surface (texture);
-	cairo_pattern_set_extend (pattern, CAIRO_EXTEND_REPEAT);
-	
 	brush->pattern = cairo_pattern_create_for_surface (texture);
+	cairo_pattern_set_extend (brush->pattern, CAIRO_EXTEND_REPEAT);
 
+	cairo_pattern_destroy (pat);
 	cairo_surface_destroy (texture);
 
 	return gdip_get_status (cairo_status (ct));
@@ -334,19 +340,19 @@ draw_clamp_texture (cairo_t *ct, GpBitmap *bitmap, GpTexture *brush)
 	/* texture surface to be created */
 	texture = cairo_surface_create_similar (original, from_cairoformat_to_content (bitmap->cairo_format),
 						rect->Width, rect->Height);
-
-	g_return_val_if_fail (texture != NULL, OutOfMemory);
-
-	cairo_save (ct);
-	{
+	if (texture == NULL) {
+		cairo_pattern_destroy (pat);
+		return OutOfMemory;
+	} else {
+		cairo_t	*ct2;
 		/* Draw the texture */
-		ct = cairo_create (texture);
-		cairo_identity_matrix (ct);
-		cairo_set_source (ct, pat);
-		cairo_rectangle (ct, 0, 0, rect->Width, rect->Height);
-		cairo_fill (ct);
+		ct2 = cairo_create (texture);
+		cairo_identity_matrix (ct2);
+		cairo_set_source (ct2, pat);
+		cairo_rectangle (ct2, 0, 0, rect->Width, rect->Height);
+		cairo_fill (ct2);
+		cairo_destroy(ct2);
 	}
-	cairo_restore (ct);
 
 	brush->pattern = cairo_pattern_create_for_surface (texture);
 
@@ -365,10 +371,10 @@ gdip_texture_setup (GpGraphics *graphics, GpBrush *brush)
 	GpImage *img, *gr_img;
 	GpBitmap *bmp, *gr_bmp;
 	cairo_format_t format;
-	GpMatrix *product = NULL;
 	unsigned int width;
 	unsigned int height;
 	GpStatus status = Ok;
+	BOOL dispose_bitmap;
 
 	g_return_val_if_fail (graphics != NULL, InvalidParameter);
 	g_return_val_if_fail (brush != NULL, InvalidParameter);
@@ -376,21 +382,25 @@ gdip_texture_setup (GpGraphics *graphics, GpBrush *brush)
 	texture = (GpTexture *) brush;
 	img = texture->image;
 	g_return_val_if_fail (img != NULL, InvalidParameter);
-	if (img->type == imageBitmap)
+	if (img->type == imageBitmap) {
 		bmp = (GpBitmap *) img;
-	else 
+		if (gdip_is_an_indexed_pixelformat (bmp->data.PixelFormat)) {
+			/* Unable to create a surface for the bitmap; it is an indexed image.
+			* Instead, it will first be converted to 32-bit RGB.
+			*/
+			bmp = gdip_convert_indexed_to_rgb (bmp);
+			g_return_val_if_fail (bmp != NULL, OutOfMemory);
+			gdip_bitmap_ensure_surface (bmp);
+			dispose_bitmap = TRUE;
+		} else {
+			dispose_bitmap = FALSE;
+		}
+		width = bmp->data.Width;
+		height = bmp->data.Height;
+		format = bmp->data.PixelFormat;
+	} else {
 		return NotImplemented;
-
-	gr_img = graphics->image;
-	g_return_val_if_fail (gr_img != NULL, InvalidParameter);
-	if (gr_img->type == imageBitmap) {
-		gr_bmp = (GpBitmap *) gr_img;
-		width = gr_bmp->data.Width;
-		height = gr_bmp->data.Height;
-		format = gr_bmp->data.PixelFormat;
 	}
-	else 
-		return NotImplemented;
 
 	ct = graphics->ct;
 	g_return_val_if_fail (ct != NULL, InvalidParameter);
@@ -399,7 +409,6 @@ gdip_texture_setup (GpGraphics *graphics, GpBrush *brush)
 	 * or if pattern has not been created yet.
 	 */
 	if (texture->base.changed || (texture->pattern) == NULL) {
-
 		/* destroy the existing pattern */
 		if (texture->pattern)
 			cairo_pattern_destroy (texture->pattern);
@@ -431,12 +440,16 @@ gdip_texture_setup (GpGraphics *graphics, GpBrush *brush)
 		}
 	}
 
-	GdipCreateMatrix (&product);
+	if (dispose_bitmap) {
+		GdipDisposeImage((GpImage *)bmp);
+	}
+
 	if (status == Ok) {
 		if (texture->pattern == NULL)
 			status = GenericError;
 		else {
-
+			GpMatrix *product = NULL;
+			GdipCreateMatrix (&product);
 		/* Cairo Bug: REPEAT and transformation do not go together.        */
 		/* To work around this we can create an intermediate surface first */
 		/* with REPEAT and use that with transformation as a pattern. But, */
@@ -479,6 +492,7 @@ gdip_texture_setup (GpGraphics *graphics, GpBrush *brush)
 			}
 			status = gdip_get_status (cairo_status (ct));
 			/*	cairo_surface_destroy (temp); */
+			GdipDeleteMatrix (product);
 		}
 	}
 
@@ -500,7 +514,6 @@ gdip_texture_clone (GpBrush *brush, GpBrush **clonedBrush)
 	texture = (GpTexture *) brush;
 	result->base = texture->base;
 	result->wrapMode = texture->wrapMode;
-	result->image = texture->image;
 
 	/* Let the clone create its own pattern. */
 	result->pattern = NULL;
@@ -516,6 +529,9 @@ gdip_texture_clone (GpBrush *brush, GpBrush **clonedBrush)
 
 	memcpy (result->rectangle, texture->rectangle, sizeof (GpRect));
 
+	result->image = texture->image;
+	cairo_surface_reference (result->image->surface);
+
 	*clonedBrush = (GpBrush *) result;
 
 	return Ok;
@@ -530,13 +546,25 @@ gdip_texture_destroy (GpBrush *brush)
 
 	texture = (GpTexture *) brush;
 
-	if (texture->rectangle != NULL)
+	if (texture->rectangle) {
 		GdipFree (texture->rectangle);
-	texture->rectangle = NULL;
+		texture->rectangle = NULL;
+	}
 
-	if (texture->pattern != NULL)
+	if (texture->pattern) {
 		cairo_pattern_destroy (texture->pattern);
-	texture->pattern = NULL;
+		texture->pattern = NULL;
+	}
+
+	if (texture->image) {
+		GdipDisposeImage (texture->image);
+		texture->image = NULL;
+	}
+
+	if (texture->matrix) {
+		GdipDeleteMatrix (texture->matrix);
+		texture->matrix = NULL;
+	}
 
 	GdipFree (texture);
 
@@ -559,21 +587,20 @@ GdipCreateTexture (GpImage *image, GpWrapMode wrapMode, GpTexture **texture)
 	imageSurface = cairo_image_surface_create_for_data ((unsigned char *)bitmap->data.Scan0,
 					bitmap->cairo_format, bitmap->data.Width,
 					bitmap->data.Height, bitmap->data.Stride);
-
 	g_return_val_if_fail (imageSurface != NULL, OutOfMemory);
 
-	/* texture surface to be used by brush */
-	image->surface = imageSurface;
-
 	*texture = gdip_texture_new ();
-
 	if (*texture == NULL) {
 		cairo_surface_destroy (imageSurface);
 		return OutOfMemory;
 	}
 
 	(*texture)->wrapMode = wrapMode;
-	(*texture)->image = image;
+	GdipCloneImage (image, &(*texture)->image);
+	// note: surface is *always* empty after cloning
+	g_assert (!(*texture)->image->surface);
+	(*texture)->image->surface = imageSurface;
+
 	(*texture)->rectangle = (GpRect *) GdipAlloc (sizeof (GpRect));
 
 	if ( (*texture)->rectangle == NULL) {
